# -*- coding: utf-8 -*-

import pytest
from .test_base_class import TestBaseClass
from .as_status_codes import AerospikeStatus
import aerospike
from aerospike import exception as e
from aerospike import predicates as p
from aerospike_helpers import expressions as exp
from aerospike_helpers import cdt_ctx
from threading import Lock
import time

list_index = "list_index"
list_rank = "list_rank"
list_value = "list_value"
map_index = "map_index"
map_key = "map_key"
map_rank = "map_rank"
map_value = "map_value"

ctx_ops = {
    list_index: cdt_ctx.cdt_ctx_list_index,
    list_rank: cdt_ctx.cdt_ctx_list_rank,
    list_value: cdt_ctx.cdt_ctx_list_value,
    map_index: cdt_ctx.cdt_ctx_map_index,
    map_key: cdt_ctx.cdt_ctx_map_key,
    map_rank: cdt_ctx.cdt_ctx_map_rank,
    map_value: cdt_ctx.cdt_ctx_map_value,
}


def add_ctx_op(ctx_type, value):
    ctx_func = ctx_ops[ctx_type]
    return ctx_func(value)


ctx_list_index = []
ctx_list_index.append(add_ctx_op(list_index, 0))

ctx_list_rank = []
ctx_list_rank.append(add_ctx_op(list_rank, -1))

ctx_list_value = []
ctx_list_value.append(add_ctx_op(list_value, 3))

ctx_map_index = []
ctx_map_index.append(add_ctx_op(map_index, 0))

ctx_map_key = []
ctx_map_key.append(add_ctx_op(map_key, "sb"))

ctx_map_rank = []
ctx_map_rank.append(add_ctx_op(map_rank, -1))

ctx_map_value = []
ctx_map_value.append(add_ctx_op(map_value, 3))


class TestQuery(TestBaseClass):
    def setup_class(cls):
        client = TestBaseClass.get_new_connection()

        try:
            client.index_integer_create("test", "demo", "test_age", "age_index")
        except e.IndexFoundError:
            pass

        try:
            client.index_string_create("test", "demo", "addr", "addr_index")
        except e.IndexFoundError:
            pass

        try:
            client.index_integer_create("test", "demo", "age1", "age_index1")
        except e.IndexFoundError:
            pass

        try:
            client.index_list_create("test", "demo", "numeric_list", aerospike.INDEX_NUMERIC, "numeric_list_index")
        except e.IndexFoundError:
            pass

        try:
            client.index_list_create("test", "demo", "string_list", aerospike.INDEX_STRING, "string_list_index")
        except e.IndexFoundError:
            pass

        try:
            client.index_map_keys_create("test", "demo", "numeric_map", aerospike.INDEX_NUMERIC, "numeric_map_index")
        except e.IndexFoundError:
            pass

        try:
            client.index_map_keys_create("test", "demo", "string_map", aerospike.INDEX_STRING, "string_map_index")
        except e.IndexFoundError:
            pass

        try:
            client.index_map_values_create(
                "test", "demo", "numeric_map", aerospike.INDEX_NUMERIC, "numeric_map_values_index"
            )
        except e.IndexFoundError:
            pass

        try:
            client.index_map_values_create(
                "test", "demo", "string_map", aerospike.INDEX_STRING, "string_map_values_index"
            )
        except e.IndexFoundError:
            pass

        try:
            client.index_integer_create("test", None, "test_age_none", "age_index_none")
        except e.IndexFoundError:
            pass

        try:
            client.index_integer_create("test", "demo", bytearray("sal\0kj", "utf-8"), "sal_index")
        except e.IndexFoundError:
            pass

        try:
            client.index_cdt_create(
                "test",
                "demo",
                "numeric_list",
                aerospike.INDEX_TYPE_DEFAULT,
                aerospike.INDEX_NUMERIC,
                "numeric_list_cdt_index",
                {"ctx": ctx_list_index},
            )
        except e.IndexFoundError:
            pass

        try:
            client.index_cdt_create(
                "test",
                "demo",
                "numeric_map",
                aerospike.INDEX_TYPE_DEFAULT,
                aerospike.INDEX_NUMERIC,
                "numeric_map_cdt_index",
                {"ctx": ctx_map_index},
            )
        except e.IndexFoundError:
            pass

        client.close()

    def teardown_class(cls):
        client = TestBaseClass.get_new_connection()

        policy = {}
        try:
            client.index_remove("test", "age_index", policy)
        except e.IndexNotFound:
            pass

        try:
            client.index_remove("test", "age_index1", policy)
        except e.IndexNotFound:
            pass

        try:
            client.index_remove("test", "addr_index", policy)
        except e.IndexNotFound:
            pass

        try:
            client.index_remove("test", "numeric_list_index", policy)
        except e.IndexNotFound:
            pass

        try:
            client.index_remove("test", "string_list_index", policy)
        except e.IndexNotFound:
            pass

        try:
            client.index_remove("test", "numeric_map_index", policy)
        except e.IndexNotFound:
            pass

        try:
            client.index_remove("test", "string_map_index", policy)
        except e.IndexNotFound:
            pass

        try:
            client.index_remove("test", "numeric_map_values_index", policy)
        except e.IndexNotFound:
            pass

        try:
            client.index_remove("test", "string_map_values_index", policy)
        except e.IndexNotFound:
            pass

        try:
            client.index_remove("test", "age_index_none", policy)
        except e.IndexNotFound:
            pass

        try:
            client.index_remove("test", "sal_index")
        except e.IndexNotFound:
            pass

        try:
            client.index_remove("test", "numeric_list_cdt_index", policy)
        except e.IndexNotFound:
            pass

        try:
            client.index_remove("test", "numeric_map_cdt_index", policy)
        except e.IndexNotFound:
            pass

        client.close()

    @pytest.fixture(autouse=True)
    def setup_method(self, request, as_connection):
        """
        Setup method.
        """
        for i in range(5):
            key = ("test", "demo", i)
            rec = {
                "name": "name%s" % (str(i)),
                "addr": "name%s" % (str(i)),
                "numeric_list": [i, i + 1, i + 2],
                "string_list": ["str" + str(i), "str" + str(i + 1), "str" + str(i + 2)],
                "numeric_map": {"a": i, "b": i + 1, "c": i + 2},
                "string_map": {"a": "a" + str(i), "b": "b" + str(i + 1), "c": "c" + str(i + 2)},
                "test_age_none": 1,
                "test_age": i,
                "no": i,
            }
            as_connection.put(key, rec)
        for i in range(5, 10):
            key = ("test", "demo", i)
            rec = {"name": "name%s" % (str(i)), "addr": "name%s" % (str(i)), "test_age": i, "no": i}
            as_connection.put(key, rec)

        key = ("test", "demo", 122)
        llist = [{"op": aerospike.OPERATOR_WRITE, "bin": bytearray("sal\0kj", "utf-8"), "val": 80000}]
        as_connection.operate(key, llist)

        key = ("test", None, 145)
        rec = {"test_age_none": 1}
        as_connection.put(key, rec)

        def teardown():
            """
            Teardown method.
            """
            for i in range(10):
                key = ("test", "demo", i)
                as_connection.remove(key)

            key = ("test", "demo", 122)
            as_connection.remove(key)
            key = ("test", None, 145)
            as_connection.remove(key)

        request.addfinalizer(teardown)

    def test_query_with_correct_parameters_hi(self):
        """
        Invoke query() with correct arguments
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.equals("test_age", 1))

        records = []

        def callback(input_tuple):
            try:
                key, _, _ = input_tuple
                records.append(key)
            except Exception as ex:
                print(ex)

        query.foreach(callback)
        assert records
        assert len(records) == 1

    def test_query_with_incorrect_bin_name(self):
        """
        Invoke query() with predicate comparing using non-extant bin name
        """
        query = self.as_connection.query("test", "demo")
        query.select("name1", "age1")
        query.where(p.equals("age1", 1))
        records = []

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.foreach(callback)
        assert len(records) == 0

    def test_query_without_callback_parameter(self):
        """
        Invoke query.foreach() with without callback
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.equals("test_age", 1))

        def callback(input_tuple):
            _, metadata, _ = input_tuple
            assert metadata["gen"] is None

        with pytest.raises(TypeError) as typeError:
            query.foreach()

        assert "argument 'callback'" in str(typeError.value)

    def test_query_with_nonindexed_bin(self):
        """
        Invoke query() with non-indexed bin
        """

        query = self.as_connection.query("test", "demo")
        query.select("name", "no")
        query.where(p.equals("no", 1))
        time.sleep(1)

        def callback(input_tuple):
            pass

        try:
            query.foreach(callback)
        except e.IndexNotFound:
            # before server version 6.0 IndexNotFound was not retryable
            assert self.server_version < [6, 0]
        except e.MaxRetriesExceeded as err_info:
            err_code = err_info.code
            # Changed to AEROSPIKE_ERR_MAX_RETRIES_EXCEEDED
            # because AEROSPIKE_ERR_INDEX_NOT_FOUND is now a retriable error.
            assert err_code == AerospikeStatus.AEROSPIKE_ERR_MAX_RETRIES_EXCEEDED
            assert "AEROSPIKE_ERR_INDEX_NOT_FOUND" in err_info.msg

    def test_query_with_where_incorrect(self):
        """
        Invoke query() with predicate which
        matches 0 records
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.equals("test_age", 165))
        records = []

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.foreach(callback)
        assert len(records) == 0

    def test_query_with_where_none_value(self):
        """
        Invoke query() with equality
        predicate comparing to None
        This is more of a predicate test

        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        try:
            query.where(p.equals("test_age", None))

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "predicate is invalid."

    def test_query_with_policy(self):
        """
        Invoke query() with policy
        """
        policy = {"timeout": 1000}
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.equals("test_age", 1))
        records = []

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.foreach(callback, policy)
        assert len(records) == 1

    @pytest.mark.xfail(reason="Query policies do not support rps")
    def test_query_with_policy_records_per_second(self):
        """
        Invoke query() with the records_per_second policy set
        """
        policy = {"timeout": 1000, "records_per_second": 10}
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.equals("test_age", 1))
        records = []

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.foreach(callback, policy)
        assert len(records) == 1

    @pytest.mark.xfail(reason="Only pass with server version >= 3.15")
    def test_query_with_no_bins_option(self):
        """
        Invoke query() with policy
        """
        policy = {"total_timeout": 1000}
        options = {"nobins": True}
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.equals("test_age", 1))

        def callback(input_tuple):
            _, _, bins = input_tuple
            print(bins)
            assert len(bins) == 0

        query.foreach(callback, policy, options=options)

    def test_query_with_invalid_options_argument_type(self):
        """
        Invoke query() with incorrect options type, should be a dictionary
        """
        policy = {"timeout": 1000}
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.equals("test_age", 1))

        def callback(input_tuple):
            _, metadata, _ = input_tuple
            assert metadata["gen"] is None

        with pytest.raises(e.ParamError):
            query.foreach(callback, policy, "")

    def test_query_with_invalid_nobins_value(self):
        """
        Invoke query() with options['nobins'] type
        """
        policy = {"total_timeout": 1000}
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.equals("test_age", 1))

        def callback(input_tuple):
            pass

        with pytest.raises(e.ParamError):
            query.foreach(callback, policy, {"nobins": "False"})

    def test_query_with_put_in_callback(self):
        """
        Invoke query() with client.put in foreach callback
        """
        policy = {"timeout": 1000}
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.equals("test_age", 1))
        records = []

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)
            key = ("test", "demo", "put_in_callback")
            rec = {
                "name": "name%s" % (str(8)),
                "test_age": 8,
            }
            self.as_connection.put(key, rec)

        query.foreach(callback, policy)

        key = ("test", "demo", "put_in_callback")
        _, _, bins = self.as_connection.get(key)
        assert bins == {"test_age": 8, "name": "name8"}

        self.as_connection.remove(key)

    def test_query_with_correct_parameters_between(self):
        """
        Invoke query() with correct arguments and using predicate between
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.between("test_age", 1, 4))

        records = []

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.foreach(callback)
        assert len(records) == 4

    def test_query_with_callback_returning_false(self):
        """
        Invoke query() with callback function returns false
        This will stop iteration
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.between("test_age", 1, 5))
        lock = Lock()
        records = []

        def callback(input_tuple):
            key, _, _ = input_tuple
            with lock:
                if len(records) == 2:
                    return False
                records.append(key)

        query.foreach(callback)
        assert len(records) == 2

    def test_query_with_results_method(self):
        """
        Invoke query() with correct arguments
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.equals("test_age", 1))

        records = query.results()
        assert len(records) == 1

    def test_query_with_results_method_and_expressions(self):
        """
        Invoke query() with correct arguments
        """
        expr = exp.Eq(exp.IntBin("test_age"), 1)

        policy = {"expressions": expr.compile()}

        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")

        records = query.results(policy)
        assert len(records) == 1

    def test_query_with_results_method_and_invalid_exp(self):
        """
        Invoke query() with correct arguments
        """
        expr = exp.Eq(exp.IntBin("test_age"), "bad_arg")

        policy = {"expressions": expr.compile()}

        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")

        with pytest.raises(e.InvalidRequest):
            query.results(policy)

    def test_query_with_results_nobins_options(self):
        """
        Invoke query() with correct arguments
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.equals("test_age", 1))

        records = query.results(options={"nobins": True})
        assert len(records) == 1

    def test_query_with_results_invalid_options_type(self):
        """
        Invoke query() with correct arguments
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.equals("test_age", 1))
        with pytest.raises(e.ParamError):
            query.results(options=False)

    def test_query_with_results_invalid_nobins_options(self):
        """
        Invoke query() with correct arguments
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.equals("test_age", 1))
        with pytest.raises(e.ParamError):
            query.results(options={"nobins": "false"})

    def test_query_with_unicode_binnames_in_select_and_where(self):
        """
        Invoke query() with unicode bin names in select
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age", "addr")
        query.where(p.equals("test_age", 7))

        records = query.results()
        assert len(records) == 1
        assert records[0][2] == {"test_age": 7, "name": "name7", "addr": "name7"}

    @pytest.mark.parametrize(
        "predicate, expected_min_length, expected_max_length",
        (
            (p.contains("numeric_list", aerospike.INDEX_TYPE_LIST, 1), 2, 2),
            (p.contains("string_list", aerospike.INDEX_TYPE_LIST, "str3"), 3, 3),
            (p.contains("string_map", aerospike.INDEX_TYPE_MAPKEYS, "a"), 5, 5),
            (p.contains("string_map", aerospike.INDEX_TYPE_MAPVALUES, "a1"), 1, 1),
            (p.contains("numeric_map", aerospike.INDEX_TYPE_MAPVALUES, 1), 2, 2),
            (p.range("numeric_map", aerospike.INDEX_TYPE_MAPVALUES, 1, 3), 4, 8),
            (p.range("numeric_list", aerospike.INDEX_TYPE_LIST, 1, 3), 4, 8),
        ),
        ids=(
            "numeric_list contains",
            "string list contains",
            "string mapkeys contains",
            "string mapvalue contains",
            "numeric mapvalue contains",
            "numeric mapvalue range",
            "numeric list range",
        ),
    )
    def test_query_with_correct_parameters_predicates(self, predicate, expected_min_length, expected_max_length):
        """
        Invoke query() with correct arguments and using predicate contains
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(predicate)

        records = []

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.foreach(callback)

        # print("TestBaseClass.major_ver:", TestBaseClass.major_ver, "TestBaseClass.minor_ver:", TestBaseClass.minor_ver)  # noqa: E501
        assert len(records) >= expected_min_length or len(records) <= expected_max_length

    def test_query_with_correct_parameters_exp(self):
        """
        Invoke query() with correct arguments and using expressions
        """

        from .test_base_class import TestBaseClass

        if TestBaseClass.major_ver >= 6 or (TestBaseClass.major_ver >= 5 and TestBaseClass.minor_ver >= 7):
            # print("TestBaseClass.major_ver:", TestBaseClass.major_ver, "TestBaseClass.minor_ver:", TestBaseClass.minor_ver)  # noqa: E501
            pytest.skip("It deprecated and it only applies to < 5.7 earlier and enterprise edition")

        expr = exp.Eq(exp.IntBin("test_age"), 4)

        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        # query.where(predicate)

        records = []

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.foreach(callback, {"expressions": expr.compile()})
        assert len(records) == 1
        assert records[0]["test_age"] == 4

    def test_query_with_multiple_foreach_on_same_query_object(self):
        """
        Invoke query() with multple foreach() call on same query object
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.equals("test_age", 1))

        records = []

        def callback(input_tuple):
            key, _, _ = input_tuple
            records.append(key)

        query.foreach(callback)
        assert len(records) == 1

        query.foreach(callback)
        assert len(records) == 2

    #  All of these removed tests feature undocumented fragile behavior removed in 3.0.0
    def test_removed_query_with_correct_parameters_rangecontains_notuple(self):
        """ """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        with pytest.raises(TypeError):
            query.where("numeric_map", "range", aerospike.INDEX_TYPE_MAPVALUES, aerospike.INDEX_NUMERIC, 1, 3)

    def test_removed_query_with_correct_parameters_containsstring_mapvalues_notuple(self):
        """
        Invoke query() with mapvalue string index
        and using predicate contains
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        with pytest.raises(TypeError):
            query.where("string_map", "contains", aerospike.INDEX_TYPE_MAPVALUES, aerospike.INDEX_STRING, "a1")

    def test_test_removed_query_containsstring_notuple(self):
        """
        Invoke query() with correct a
        and a string list index arguments and using predicate contains
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        with pytest.raises(TypeError):
            query.where("string_list", "contains", aerospike.INDEX_TYPE_LIST, aerospike.INDEX_STRING, "str3")

    def test_removed_query_with_correct_parameters_between_notuple(self):
        """
        Invoke query() with correct arguments and using predicate between
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        with pytest.raises(TypeError):
            query.where("test_age", "between", 1, 4)

    def test_between_predicate_between_one_arg(self):
        """
        Invoke query and using predicate between with invalid predicate
        arguments
        """
        query = self.as_connection.query("test", "demo")
        with pytest.raises(TypeError):
            query.where("test_age", "between", 1)

    def test_between_predicate_between_no_args(self):
        """
        Invoke query and using predicate between with invalid predicate
        arguments
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        with pytest.raises(e.ParamError):
            query.where("test_age", "between")

    # TODO: incorrect test
    def test_query_with_policy_notuple(self):
        """
        Invoke query() with policy
        """
        policy = {"timeout": 1000}  # noqa: F841
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        with pytest.raises(TypeError):
            query.where("test_age", "equals", 1)

    def test_query_with_multiple_results_call_on_same_query_object(self):
        """
        Invoke query() with multple results() call on same query object
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age", "addr")
        query.where(p.equals("test_age", 7))

        records = query.results()
        assert len(records) == 1
        assert records[0][2] == {"test_age": 7, "name": "name7", "addr": "name7"}

        records = []
        records = query.results()
        assert len(records) == 1
        assert records[0][2] == {"test_age": 7, "name": "name7", "addr": "name7"}

    def test_query_with_policy_on_none_set_index(self):
        """
        Invoke query() with set specified as None
        """
        policy = {"timeout": 1000}
        query = self.as_connection.query("test", None)
        query.select("name", "test_age_none")
        query.where(p.equals("test_age_none", 1))
        records = []

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        # This should only select records in the ns with no set
        # that match the where clause, if it matched all in the ns,
        # this would return 6
        query.foreach(callback, policy)
        if TestBaseClass.major_ver < 6 or (TestBaseClass.major_ver == 6 and TestBaseClass.minor_ver == 0):
            assert len(records) == 1
        else:
            assert len(records) == 6

    def test_query_with_only_ns(self):
        """
        Invoke query() with only ns specified
        """
        policy = {"timeout": 1000}
        query = self.as_connection.query("test")
        query.select("name", "test_age_none")
        query.where(p.equals("test_age_none", 1))
        records = []

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        # This should only select records in the ns with no set
        # that match the where clause, if it matched all in the ns,
        # this would return 6
        query.foreach(callback, policy)
        if TestBaseClass.major_ver < 6 or (TestBaseClass.major_ver == 6 and TestBaseClass.minor_ver == 0):
            assert len(records) == 1
        else:
            assert len(records) == 6

    def test_query_with_select_bytearray(self):
        """
        Invoke query() with correct arguments
        and select and equality predicate comparing a byte array
        named bin
        """
        query = self.as_connection.query("test", "demo")
        query.select(bytearray("sal\0kj", "utf-8"))
        query.where(p.equals(bytearray("sal\0kj", "utf-8"), 80000))

        records = []

        def callback(input_tuple):
            key, _, _ = input_tuple
            records.append(key)

        query.foreach(callback)
        assert records
        assert len(records) == 1

    def test_query_with_no_parameters(self):
        """
        Invoke query() without any mandatory parameters.
        """
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.query()

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_query_with_incorrect_policy_value(self):
        """
        Invoke query() with invalid policy passed to foreach
        """
        policy = {"total_timeout": ""}
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.equals("test_age", 1))

        def callback(input_tuple):
            _, metadata, _ = input_tuple
            print(metadata)
            assert metadata["gen"] is None

        with pytest.raises(e.ParamError) as err_info:
            query.foreach(callback, policy)

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_query_with_select_bin_integer(self):
        """
        Invoke query() with select bin is of integer type.
        Try to select a bin with an integer name instead of a string
        """
        query = self.as_connection.query("test", "demo")

        with pytest.raises(e.ParamError) as err_info:
            query.select(22, "test_age")

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_query_with_argument_to_where_is_empty_string(self):

        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")

        with pytest.raises(e.ParamError) as err_info:
            query.where("")

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_query_with_callback_contains_error(self):
        """
        Invoke query() with callback function contains an error
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.equals("test_age", 1))

        def callback(input_tuple):
            raise Exception("error")

        with pytest.raises(e.ClientError) as err_info:
            query.foreach(callback)

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_CLIENT

    def test_query_with_incorrect_ns_set(self):
        """
        Invoke query() with incorrect ns and set
        foreach should raise an error since the ns
        is invalid
        """
        with pytest.raises(e.ClientError) as err_info:
            query = self.as_connection.query("fake_namespace", "demo1")

            def callback(input_tuple):
                """
                no-op callback
                """

            query.foreach(callback)

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_CLIENT

    def test_query_with_ns_not_string(self):
        """
        Invoke query() with incorrect ns and set. It
        should raises a param error.
        """
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.query(1, "demo")

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_query_with_set_int(self):
        """
        Invoke query() with set argument not a string
        """
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.query("test", 1)

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_query_with_correct_parameters_without_connection(self):
        """
        Invoke query() with correct arguments without connection
        """
        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()

        with pytest.raises(e.ClusterError) as err_info:
            query = client1.query("test", "demo")
            query.select("name", "test_age")
            query.where(p.equals("test_age", 1))

            def callback(input_tuple):
                pass

            query.foreach(callback)

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_CLUSTER_ERROR

    @pytest.mark.skip(reason="segfault")
    def test_query_predicate_range_wrong_no_args(self):
        """ """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where("numeric_map", "range", aerospike.INDEX_TYPE_MAPVALUES, aerospike.INDEX_NUMERIC)

    @pytest.mark.skip(reason="segfault")
    def test_query_predicate_range_wrong_one_end_args(self):
        """ """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where("numeric_map", "range", aerospike.INDEX_TYPE_MAPVALUES, aerospike.INDEX_NUMERIC, 1)

    def test_query_with_list_cdt_ctx(self):
        """
        Invoke query() with cdt_ctx and correct arguments
        """
        from .test_base_class import TestBaseClass

        if TestBaseClass.major_ver < 6 or (TestBaseClass.major_ver == 6 and TestBaseClass.minor_ver == 0):
            pytest.skip("It only applies to >= 6.1 enterprise edition")

        # ctx = []
        # ctx.append(add_ctx_op(list_index, 0))
        # query = self.as_connection.query('test', 'demo')
        # query.select('numeric_list')
        # query.where(p.range('numeric_list', aerospike.INDEX_TYPE_DEFAULT, 2,4), {'ctx':ctx})

        query = self.as_connection.query("test", "demo")
        query.select("numeric_list")
        query.where(p.range("numeric_list", aerospike.INDEX_TYPE_DEFAULT, 2, 4), {"ctx": ctx_list_index})

        records = []

        def callback(input_tuple):
            try:
                records.append(input_tuple)
            except Exception as ex:
                print(ex)

        query.foreach(callback)

        assert records
        assert len(records) == 3

    def test_query_with_map_cdt_ctx(self):
        """
        Invoke query() with cdt_ctx and correct arguments
        """
        from .test_base_class import TestBaseClass

        if TestBaseClass.major_ver < 6 or (TestBaseClass.major_ver == 6 and TestBaseClass.minor_ver == 0):
            pytest.skip("It only applies to >= 6.1 enterprise edition")

        query = self.as_connection.query("test", "demo")
        query.select("numeric_map")

        query.where(p.range("numeric_map", aerospike.INDEX_TYPE_DEFAULT, 2, 4), {"ctx": ctx_map_index})

        records = []

        def callback(input_tuple):
            try:
                records.append(input_tuple)
            except Exception as ex:
                print(ex)

        query.foreach(callback)

        assert records
        assert len(records) == 3

    def test_query_with_base64_cdt_ctx(self):
        bs_b4_cdt = self.as_connection.get_cdtctx_base64({"ctx": ctx_list_index})
        assert bs_b4_cdt == "khAA"



# -*- coding: utf-8 -*-
################################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

from __future__ import print_function

import aerospike
import sys

from optparse import OptionParser

################################################################################
# Options Parsing
################################################################################

usage = "usage: %prog [options]"

optparser = OptionParser(usage=usage, add_help_option=False)

optparser.add_option(
    "--help", dest="help", action="store_true",
    help="Displays this message.")

optparser.add_option(
    "-h", "--host", dest="host", type="string", default="127.0.0.1", metavar="<ADDRESS>",
    help="Address of Aerospike server.")

optparser.add_option(
    "-p", "--port", dest="port", type="int", default=3000, metavar="<PORT>",
    help="Port of the Aerospike server.")

optparser.add_option(
    "-U", "--username", dest="username", type="string", metavar="<USERNAME>",
    help="Username to connect to database.")

optparser.add_option(
    "-P", "--password", dest="password", type="string", metavar="<PASSWORD>",
    help="Password to connect to database.")

(options, args) = optparser.parse_args()

if options.help:
    optparser.print_help()
    print()
    sys.exit(1)

if options.username == None or options.password == None:
    optparser.print_help()
    print()
    sys.exit(1)

################################################################################
# Client Configuration
################################################################################

config = {
    'hosts': [ (options.host, options.port) ]
}

################################################################################
# Application
################################################################################

exitCode = 0

try:

    # ----------------------------------------------------------------------------
    # Connect to Cluster
    # ----------------------------------------------------------------------------

    client = aerospike.client(config).connect(options.username, options.password)

    # ----------------------------------------------------------------------------
    # Perform Operation
    # ----------------------------------------------------------------------------
     
    try:

        policy = {}
        role = "example_foo"
    	
        privileges = client.admin_query_role(role, policy)
        
        print(privileges)
        print("---")
        print("OK, Privileges retrieved for one role")

    except Exception as e:
        print("error: {0}".format(e), file=sys.stderr)
        print("In case of invalid role, first create role using create_role.py")
        exitCode = 2
    
    # ----------------------------------------------------------------------------
    # Close Connection to Cluster
    # ----------------------------------------------------------------------------

    client.close()

except Exception as eargs:
    print("error: {0}".format(eargs), file=sys.stderr)
    exitCode = 3

################################################################################
# Exit
################################################################################

sys.exit(exitCode)



# -*- coding: utf-8 -*-
##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################

from __future__ import print_function

import aerospike
import random
import signal
import sys
import string
import time


from optparse import OptionParser
import copy
from tabulate import tabulate

# Guppy is only available in python 2.7, try to load it
try:
    from guppy import hpy
    HAVE_HEAPY = True
except ImportError:
    HAVE_HEAPY = False
# set maximum integer value compatible in python2 and 3
try:
    MAX_INT = sys.maxint
except AttributeError:
    MAX_INT = sys.maxsize

##########################################################################
# Options Parsing
##########################################################################

"""
TODO: change optparse to argparse
optparse is deprecated from Python V2.7
"""
usage = "usage: %prog [options]"

optparser = OptionParser(usage=usage, add_help_option=False)

optparser.add_option(
    "--help", dest="help", action="store_true",
    help="Displays this message.")

optparser.add_option(
    "-U", "--username", dest="username", type="string", metavar="<USERNAME>",
    help="Username to connect to database.")

optparser.add_option(
    "-P", "--password", dest="password", type="string", metavar="<PASSWORD>",
    help="Password to connect to database.")

optparser.add_option(
    "-h", "--host", dest="host", type="string", default="127.0.0.1", metavar="<ADDRESS>",
    help="Address of Aerospike server.")

optparser.add_option(
    "-p", "--port", dest="port", type="int", default=3000, metavar="<PORT>",
    help="Port of the Aerospike server.")

optparser.add_option(
    "-n", "--namespace", dest="namespace", type="string", default="test", metavar="<NS>",
    help="Namespace that records will be stored and retrieved from.")

optparser.add_option(
    "-s", "--set", dest="set", type="string", default="demo", metavar="<SET>",
    help="Set that records will be stored and retrieved from.")

optparser.add_option(
    "-v", "--verbose", dest="verbose", action="store_true", metavar="<PORT>",
    help="Verbose output.")

optparser.add_option(
    "--gen", dest="gen", type="string", default="int",
    help="Key and value generator: int | str")

optparser.add_option(
    "--str-min", dest="str_min", type="int", default=10,
    help="Minimum length for generated strings.")

optparser.add_option(
    "--str-max", dest="str_max", type="int", default=30,
    help="Maximum length for generated strings.")

optparser.add_option(
    "--str-chars", dest="str_chars", type="string", default=string.ascii_uppercase + string.digits,
    help="Charset for generated strings.")

optparser.add_option(
    "--int-min", dest="int_min", type="int", default=0,
    help="Minimum value for generated integers.")

optparser.add_option(
    "--int-max", dest="int_max", type="int", default=MAX_INT,
    help="Maximum value for generated integere.")

optparser.add_option(
    "--heap", dest="heap", action="store_true",
    help="Check the heap. Produce a heap report at the end of the run.")

optparser.add_option(
    "--heap-interval", dest="heap_interval", type="int", default=0,
    help="Display heap report after every n operations.")

optparser.add_option(
    "--reads", dest="reads", type="int", default=80,
    help="Read ratio, as an integer between 0 and 100")

optparser.add_option(
    "--writes", dest="writes", type="int", default=20,
    help="Write ratio, as an integer between 0 and 100")

optparser.add_option(
    "--keys", dest="keys", type="int", default=1000000,
    help="Number of unique keys")

optparser.add_option(
    "-l", "--latency", dest="latency", type="string", default="4,1",
    help="<number of latency columns>,<range shift increment> Show transaction latency percentages using elapsed time ranges.\
    <number of latency columns>: Number of elapsed time ranges. <range shift increment>: Power of 2 multiple between each range starting at column 3.\
    A latency definition of '-latency 7,1' results in this layout:\
        <=1ms >1ms >2ms >4ms >8ms >16ms >32ms\
           x%   x%   x%   x%   x%    x%    x%\
    A latency definition of '-latency 4,3' results in this layout:\
        <=1ms >1ms >8ms >64ms\
           x%   x%   x%    x%\
    Latency columns are cumulative. If a transaction takes 9ms, it will be included in both the >1ms and >8ms columns.")

(options, args) = optparser.parse_args()

if options.help:
    optparser.print_help()
    print()
    sys.exit(1)

##########################################################################
# Client Configuration
##########################################################################

config = {
    'hosts': [(options.host, options.port)]
}

##########################################################################
# Generators
##########################################################################

READ_OP = 0
WRITE_OP = 1
CHOICE_OP = [READ_OP, WRITE_OP]

# Generator for operations


def operation(r, w):
    rn = r
    wn = w
    total = r + w
    while rn + wn > 0:
        if rn > 0 and wn > 0:
            op = random.randint(0, 100)
            if op <= r:
                op = READ_OP
                rn -= 1
            else:
                op = WRITE_OP
                wn -= 1
            if wn == 0 and rn == 0:
                rn = r
                wn = w
            yield op
        elif rn > 0:
            op = READ_OP
            rn -= 1
            if wn == 0 and rn == 0:
                rn = r
                wn = w
            yield op
        elif wn > 0:
            op = WRITE_OP
            wn -= 1
            if wn == 0 and rn == 0:
                rn = r
                wn = w
            yield op


def genstr():
    return ''.join(random.choice(options.str_chars) for _ in range(random.randrange(options.str_min, options.str_max, 1)))


def genint():
    return random.randrange(options.int_min, options.int_max, 1)


##########################################################################
# Application
##########################################################################

exitCode = 0

k = options.keys
r = options.reads
w = options.writes

total_count = 0
read_count = 0
write_count = 0
start = 0
intervals = []
if HAVE_HEAPY:
    heapy = hpy()

no_of_buckets, range_increment = [
    int(elem.strip()) for elem in options.latency.split(',')]

__buckets = [0, 1] + [elem << i for i in [0]
                      for elem in [1] * (no_of_buckets - 2) for i in [i + range_increment]]

# Separate buckets for read / write latency data
read_bucket = dict.fromkeys(__buckets, 0)
write_bucket = copy.deepcopy(read_bucket)

# return current time in ms
current_milliseconds_time = lambda: int(round(time.time() * 1000))

# returns latency table headers


def get_latency_table_headers():
    headers = ["", "<=1ms", ">1ms"] + \
        [">{0}ms".format(i) for i in __buckets[2:]]
    return headers

# manages the read / write operations counter in latency dictionaries


def increment_counters(bucket, time_in_millisecond):
    if time_in_millisecond <= 0:
        bucket[0] += 1
    for key in bucket.keys():
        if time_in_millisecond >= key > 0:
            bucket[key] += 1

# format the output as per tabulate standard and calculate percentages as well


def interprete_summary():
    global total_count, read_count, write_count, read_bucket, write_bucket

    read_summary = ['read'] + ['{0}%'.format(
        int(round(read_bucket[key] * 100 / read_count))) for key in read_bucket.keys()]
    write_summary = ['write'] + ['{0}%'.format(
        int(round(write_bucket[key] * 100 / write_count))) for key in write_bucket.keys()]

    return read_summary, write_summary


def total_summary():

    # stop time
    stop = time.time()

    # elapse time
    elapse = (stop - start)

    print()
    print("Summary:")
    print("     {0} keys generated".format(k))
    print("     {0} seconds for {1} operations".format(elapse, total_count))
    print("     {0} operations per second".format(total_count / elapse))
    print()
    print("Latency stats:")
    table = [info for info in interprete_summary()]
    print(tabulate(table, headers=get_latency_table_headers()))
    print()
    if HAVE_HEAPY:
        print("Heap: ")
        print(heapy.heap())

    sys.exit(0)

try:

    # ----------------------------------------------------------------------------
    # Generate Keys
    # ----------------------------------------------------------------------------

    if options.gen == 'str':
        gen = genstr
    else:
        gen = genint

    print("Generating keys: ", k)
    keys = [gen() for _ in range(k)]
    print("done.")
    print()

    # ----------------------------------------------------------------------------
    # Connect to Cluster
    # ----------------------------------------------------------------------------

    client = aerospike.client(config).connect(
        options.username, options.password)

    # ----------------------------------------------------------------------------
    # Perform Operation
    # ----------------------------------------------------------------------------

    try:

        signal.signal(signal.SIGTERM, total_summary)

        print()
        print("Press CTRL+C to quit.")
        print()

        start = time.time()

        if HAVE_HEAPY:
            print("HEAP@{0}: {1}".format(0, heapy.heap()))

        # run the operatons
        for op in operation(r, w):

            key = (options.namespace, options.set,
                   keys[random.randint(0, k - 1)])

            if op == READ_OP:
                print('[READ] ', key) if options.verbose else 0
                now = current_milliseconds_time()
                result = client.exists(key)
                elapsed = current_milliseconds_time() - now
                print('[READ ELAPSED] {0}ms'.format(
                    elapsed)) if options.verbose else 0
                increment_counters(read_bucket, elapsed)
                total_count += 1
                read_count += 1

            elif op == WRITE_OP:
                print('[WRITE]', key) if options.verbose else 0
                rec = {
                    'key': key[2]
                }
                now = current_milliseconds_time()
                client.put(key, rec)
                elapsed = current_milliseconds_time() - now
                print('[WRITE ELAPSED] {0}ms'.format(
                    elapsed)) if options.verbose else 0
                increment_counters(write_bucket, elapsed)
                total_count += 1
                write_count += 1

            if options.heap_interval > 0 and (total_count % options.heap_interval) == 0:
                if HAVE_HEAPY:
                    print("HEAP@{0}: {1}".format(total_count, heapy.heap()))

    except KeyboardInterrupt:
        total_summary()
    except Exception as eargs:
        print("error: {0}".format(eargs), file=sys.stderr)
        sys.exit(2)

except Exception as eargs:
    print("error: {0}".format(eargs), file=sys.stderr)
    sys.exit(3)

##########################################################################
# Exit
##########################################################################

sys.exit(0)



# -*- coding: utf-8 -*-
##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################


# Test the TTL feature for the Aerospike Server.
# We will write records that expire BEFORE the default namespace TTL,
# we will write records that expire with the default TTL, and we'll
# write records that NEVER expire.
#
# NOTE:
# This test is meant to run on an Aerospike 2.x or 3.x server, so the
# records that it writes have only primitive types for bin values.

from __future__ import print_function

import aerospike
import sys
import textwrap
import re
import time

from optparse import OptionParser
from aerospike import exception as e

##########################################################################
# Option Parsing
##########################################################################

usage = "usage: %prog [options]"

optparser = OptionParser(usage=usage, add_help_option=False)

optparser.add_option(
    "--help", dest="help", action="store_true",
    help="Displays this message.")

optparser.add_option(
    "-U", "--username", dest="username", type="string", metavar="<USERNAME>",
    help="Username to connect to database.")

optparser.add_option(
    "-P", "--password", dest="password", type="string", metavar="<PASSWORD>",
    help="Password to connect to database.")

optparser.add_option(
    "-h", "--host", dest="host", type="string", default="127.0.0.1", metavar="<ADDRESS>",
    help="Address of Aerospike server.")

optparser.add_option(
    "-p", "--port", dest="port", type="int", default=3000, metavar="<PORT>",
    help="Port of the Aerospike server.")

optparser.add_option(
    "-n", "--namespace", dest="namespace", type="string", default="test", metavar="<NS>",
    help="Port of the Aerospike server.")

optparser.add_option(
    "-s", "--set", dest="set", type="string", default="demo", metavar="<SET>",
    help="Port of the Aerospike server.")

(options, args) = optparser.parse_args()

if options.help:
    optparser.print_help()
    print()
    sys.exit(1)

##########################################################################
# CONSTANTS
##########################################################################

TTL_DEFAULT = 10
TTL_MAX = 20
TTL_NO_EXPIRE = -1


# Define the Namespace Supervisor parms -- setting the period very short
# so that we know it will have visited all of our records before we look
# at them at each TTL interval.
PARAMS_SERVICE = [[('nsup-period', 1)]]

# Define the default Namespace Time To Live at 10 seconds. We will write
# some records that expire EARLY (5 seconds), some records that expire at
# the default (10 seconds), some that expire LATE (15 seconds) and some
# that NEVER expire.
# We set MAX ttl to 20 to check that our flag (0xFFFFFFFF) is allowed in
# and does not trigger the "greater than max ttl" warning.
# Also, we'll check that one of our records DOES trigger the Max TTL warning
# with a TTL of greater than 20.
PARAMS_NAMESPACE = [[('default-ttl', TTL_DEFAULT), ('max-ttl', TTL_MAX)]]

# Write Policy names and related values
AS_POLICY_W_TIMEOUT = "timeout"

AS_POLICY_W_RETRY = "retry"
AS_POLICY_RETRY_UNDEF = 0  # Use Default value
AS_POLICY_RETRY_NONE = 1  # No retry
AS_POLICY_RETRY_ONCE = 2  # Retry Once

AS_POLICY_W_KEY = "key"
AS_POLICY_KEY_UNDEF = 0  # If set, then the value will default to either
# as_config.policies.key or `AS_POLICY_KEY_DEFAULT`.
AS_POLICY_KEY_DIGEST = 1  # Send the digest value of the key.
AS_POLICY_KEY_SEND = 2  # Send the key, but do not store it.
AS_POLICY_KEY_STORE = 3  # Store the key (NOT YET IMPLEMENTED)

AS_POLICY_W_GEN = "generation"
AS_POLICY_GEN_UNDEF = 0  # Use default value
AS_POLICY_GEN_IGNORE = 1  # Write a record, regardless of generation.
AS_POLICY_GEN_EQ = 2  # Write a record, ONLY if generations are equal
AS_POLICY_GEN_GT = 3  # Write a record, ONLY if local generation is
# greater-than remote generation.
AS_POLICY_GEN_DUP = 4  # Write a record creating a duplicate, ONLY if
# the generation collides (?)

AS_POLICY_W_EXISTS = "exists"
AS_POLICY_EXISTS_UNDEF = 0  # Use default value
AS_POLICY_EXISTS_IGNORE = 1  # Write the record, regardless of existence.
AS_POLICY_EXISTS_CREATE = 2  # Create a record, ONLY if it doesn't exist.
# Update a record, ONLY if it exist (NOT YET IMPL).
AS_POLICY_EXISTS_UPDATE = 3

# Setup write policy
wr_policy = {
    AS_POLICY_W_TIMEOUT: 5000,
    AS_POLICY_W_RETRY:   AS_POLICY_RETRY_NONE,
    AS_POLICY_W_KEY:     AS_POLICY_KEY_DIGEST,
    AS_POLICY_W_GEN:     AS_POLICY_GEN_IGNORE,
    AS_POLICY_W_EXISTS:  AS_POLICY_EXISTS_IGNORE
}

BASE_KEY_RANGE = list(range(1, 11))

SPECIAL_KEYS = {
    20: {'ttl': 5,
         'desc': '5 sec TTL'},
    40: {'ttl': 15,
         'desc': '15 sec TTL'},
    60: {'ttl': TTL_NO_EXPIRE,
         'desc': 'NO_EXPIRE TTL'},
    80: {'ttl': TTL_MAX + 1,
         'desc': 'Larger than MAX TTL'}
}

KEYS = BASE_KEY_RANGE + list(SPECIAL_KEYS.keys())

##########################################################################
# Connect to Cluster
##########################################################################

config = {
    'hosts': [(options.host, options.port)]
}

print('Connect to Server: ', config)
client = aerospike.client(config).connect(options.username, options.password)

##########################################################################
# Perform Operation
##########################################################################


def test_params_for_stanza(p, contx, is_namespace):
    for t in p:
        if is_namespace:
            info_code = "set-config:context=namespace;id=" + \
                contx + ";" + t[0] + "=" + str(t[1])
        else:
            info_code = "set-config:context=" + \
                contx + ";" + t[0] + "=" + str(t[1])
        try:
            v = list(client.info(info_code).items())
            res = v[0][1][1]
            if res != "ok\n":
                print(
                    "setting {0} to {1} failed. Got result({2})".format(t[0], t[1], res))
                sys.exit(1)
            else:
                print("setting {0} to {1} succeeded".format(t[0], t[1]))
        except:
            print("error: info_code({0}) result({1})".format(info_code, v))


def print_header(header, message=None):
    print()
    print(''.ljust(80, '='))
    print(header)
    print(message) if message else None
    print(''.ljust(80, '-'))


def print_record(xxx_todo_changeme, prefix=''):
    (key, meta, record) = xxx_todo_changeme
    print("%s%-4d %-4s %-8s %s" % (
        prefix,
        int(key[2] or 0),
        meta.get('gen') if meta and 'gen' in meta else '-',
        meta.get('ttl') if meta and 'ttl' in meta else '-',
        record if record else '-'
    ))


def print_records(records, prefix=''):
    header = "%s---- ---- -------- " % prefix
    header = header.ljust(80, '-')
    print()
    print("%s%-4s %-4s %-8s %s" % (prefix, "key", "gen", "ttl", "record"))
    print(header)
    [print_record(r, prefix) for r in records]


def print_histogram(prefix=''):
    request = ''.join(["hist-dump:ns=", options.namespace, ";hist=ttl"])

    header = "%sHISTOGRAM (%s)" % (prefix, request)
    border = prefix.ljust(80, '-')

    print()
    print(header)
    print(border)
    for _, (error, response) in list(client.info(request).items()):
        if error:
            print('%serror: %s' % (prefix, error))
        else:
            for line in textwrap.wrap(response, 80 - len(prefix)):
                print("%s%s" % (prefix, line))


def check_records(start, wait=0, message=None):

    if wait:
        time.sleep(wait)

    stop = time.time()
    duration = int(stop - start)

    print_header('CHECK :: wait=%s duration=%s' % (wait, duration), message)

    try:
        print_records(
            [client.get((options.namespace, options.set, k)) for k in KEYS], '  ')
    except Exception as e:
        print("error: {0}".format(e), file=sys.stderr)

    print_histogram('  ')


def delete_records():
    try:
        for key in KEYS:
            # first remove the existing record
            client.remove((options.namespace, options.set, key))
    except e.RecordNotFound:
        print("Record not found")
    except Exception as err:
        if err[0] != 2:
            print("delete_records() error: {0}".format(
                err[0]), file=sys.stderr)
            sys.exit(1)


def write_records():

    try:
        for key in KEYS:
            ttl = SPECIAL_KEYS[key]['ttl'] if key in SPECIAL_KEYS else None

            rec = {}
            rec['key'] = key
            rec['ttl'] = ttl if ttl else TTL_DEFAULT
            rec['desc'] = SPECIAL_KEYS[key][
                'desc'] if key in SPECIAL_KEYS else 'default TTL'

            try:
                # write a new record
                # ttl=None is equivalent to not setting a ttl
                print("writing key :=", key)
                client.put(
                    (options.namespace, options.set, key), rec, {'ttl': ttl})

            except Exception as e:
                ttlVal = int(ttl or 0)
                if ttlVal > TTL_MAX:
                    print('error: (correct) failed to write record with TTL(%d) > TTL_MAX(%d)' % (
                        ttlVal, TTL_MAX))
                else:
                    print('error: failed to write record with TTL = %d ' %
                          ttlVal)

    except Exception as e:
        print("error: {0}".format(e), file=sys.stderr)
        sys.exit(1)

##########################################################################
# CONFIGURE SERVER
##########################################################################

print_header("CONFIGURE THE SERVER")

# Now go off and set the params
print('Set Parameters for Service')
for p in PARAMS_SERVICE:
    test_params_for_stanza(p, "service", False)
    time.sleep(1)
print("service parameters passed")

print("getting initial ttl values")
info = client.info("namespace/" + options.namespace)
default_ttl = 0
max_ttl = 0
for key, value in list(info.items()):
    array_of_items = re.split(';|=', value[1])
    i = 0
    for item in array_of_items:
        i = i + 1
        if item == "default-ttl":
            default_ttl = array_of_items[i]
        if item == "max-ttl":
            max_ttl = array_of_items[i]

print('Set Parameters for Namespace')
for p in PARAMS_NAMESPACE:
    test_params_for_stanza(p, options.namespace, True)
    time.sleep(1)
print("namespace parameters passed")

##########################################################################
# CHECK RECORDS ON INTERVALS
##########################################################################

start = time.time()

delete_records()
check_records(start, 0, 'Clean state')

write_records()

check_records(start, 0, 'Initial state')
check_records(start, 2, 'Expect all records with TTL-2')
check_records(start, 6, 'Expect all records with TTL<=5 to be gone')
check_records(start, 3, 'Expect all records with TTL<=10 to be gone')
check_records(start, 6, 'Expect all records to be gone, except NO_EXPIRE')
client.remove((options.namespace, options.set, 60))

PARAMS_NAMESPACE = [[('default-ttl', default_ttl), ('max-ttl', max_ttl)]]
print('Reset Parameters for Namespace')
for p in PARAMS_NAMESPACE:
    test_params_for_stanza(p, options.namespace, True)
    time.sleep(1)



# -*- coding: utf-8 -*-

import pytest
from .test_base_class import TestBaseClass

try:
    from collections import Counter
except ImportError:
    from counter26 import Counter

import aerospike
from aerospike_helpers.operations import operations


class TestCompress:

    config = TestBaseClass.get_connection_config()

    pytestmark = pytest.mark.skipif(
        not TestBaseClass.enterprise_in_use(), reason="No enterprise hosts, maybe cluster is not enterprise."
    )

    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        self.keys = []

        for i in range(5):
            key = ("test", "demo", i)
            rec = {"name": "name%s" % (str(i)), "age": i}
            as_connection.put(key, rec)
            self.keys.append(key)

        key = ("test", "demo", "float_value")
        as_connection.put(key, {"float_value": 4.3})
        self.keys.append(key)

        def teardown():
            """
            Teardown method.
            """
            for i in range(5):
                key = ("test", "demo", i)
                as_connection.remove(key)

            key = ("test", "demo", "float_value")
            as_connection.remove(key)

        request.addfinalizer(teardown)

    def test_put_get_with_compress_policy(self):
        """
        Invoke put() and get() with compression policy.
        """
        key = ("test", "demo", 1)
        expected = "john" * (5 * 1024)
        rec = {"val": expected}

        policy = {"compress": True}

        self.as_connection.put(key, rec, policy)
        _, _, bins = self.as_connection.get(key, policy)

        assert bins["val"] == expected

    def test_batch_with_compress_policy(self):
        """
        Invoke get_many() with compression policy.
        """
        policy = {"compress": True}
        records = self.as_connection.get_many(self.keys, policy)

        assert isinstance(records, list)
        assert len(records) == 6
        assert Counter([x[0][2] for x in records]) == Counter([0, 1, 2, 3, 4, "float_value"])
        assert records[5][2] == {"float_value": 4.3}

    def test_operate_with_compress_policy(self):
        """
        Invoke operate() with compression policy.
        """
        key = ("test", "demo", 1)
        policy = {
            "timeout": 1000,
            "key": aerospike.POLICY_KEY_SEND,
            "commit_level": aerospike.POLICY_COMMIT_LEVEL_MASTER,
            "compress": True,
        }

        llist = [
            {"op": aerospike.OPERATOR_APPEND, "bin": "name", "val": "aa"},
            {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
            {"op": aerospike.OPERATOR_READ, "bin": "name"},
        ]

        key, _, bins = self.as_connection.operate(key, llist, {}, policy)

        assert bins == {"name": "name1aa"}
        assert key == (
            "test",
            "demo",
            1,
            bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
        )

    def test_read_with_compress_policy(self, put_data):
        """
        Invoke get() for a record with compression policy.
        """
        key = ("test", "demo", 1)
        rec = {"name": "john", "age": 1}

        policy = {"key": aerospike.POLICY_KEY_SEND, "compress": True}

        self.as_connection.put(key, rec, policy)

        key, _, bins = self.as_connection.get(key, policy)

        assert bins == {"name": "john", "age": 1}
        assert key == (
            "test",
            "demo",
            1,
            bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
        )

    def test_scan_with_compress_policy(self):
        """
        Invoke execute_background() for a scan with compression enabled.
        """

        ops = [operations.touch()]

        scan_obj = self.as_connection.scan("test", "demo")
        scan_obj.add_ops(ops)
        scan_obj.execute_background({"compress": True})



# -*- coding: utf-8 -*-

import math

import pytest


from aerospike_helpers.expressions import (
    Abs,
    Add,
    Ceil,
    Div,
    Eq,
    FloatBin,
    Floor,
    IntBin,
    Log,
    Max,
    Min,
    Mod,
    Mul,
    Pow,
    StrBin,
    Sub,
    ToFloat,
    ToInt,
)
from aerospike import exception as e
from .test_base_class import TestBaseClass


class TestExpressionsArithmetic(TestBaseClass):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        if self.server_version < [5, 6]:
            pytest.mark.xfail(reason="Servers older than 5.6 do not support arithmetic expressions.")
            pytest.xfail()

        self.test_ns = "test"
        self.test_set = "demo"

        self.key = ("test", "demo", 5)
        self.rec = {"name": "name5", "fbin": 5.0, "ibin": 5}
        self.as_connection.put(self.key, self.rec)

        def teardown():
            as_connection.remove(self.key)

        request.addfinalizer(teardown)

    def verify_expression(self, expr, expected):
        _, _, res = self.as_connection.get(self.key, policy={"expressions": expr})
        assert res == self.rec

    def verify_expression_neg(self, expr, expected):
        with pytest.raises(expected):
            _, _, res = self.as_connection.get(self.key, policy={"expressions": expr})

    @pytest.mark.parametrize(
        "bin, val, check", [(IntBin("ibin"), [5], 10), (IntBin("ibin"), [5, 100], 110), (FloatBin("fbin"), [3.0], 8.0)]
    )
    def test_add_pos(self, bin, val, check):
        """
        Test arithmetic Add expression with correct parameters.
        """
        expr = Eq(Add(bin, *val), check).compile()

        self.verify_expression(expr, self.rec)

    @pytest.mark.parametrize(
        "bin, val, check, expected",
        [
            (IntBin("ibin"), [5], 25, e.FilteredOut),
            (IntBin("ibin"), [5.0], 10, e.InvalidRequest),
            (FloatBin("fbin"), [3], 8.0, e.InvalidRequest),
            (FloatBin("fbin"), ["bad_arg"], 8.0, e.InvalidRequest),
        ],
    )
    def test_add_neg(self, bin, val, check, expected):
        """
        Test arithmetic Add expression expecting failure.
        """
        expr = Eq(Add(bin, *val), check).compile()

        self.verify_expression_neg(expr, expected)

    @pytest.mark.parametrize(
        "expression",
        [
            Eq(IntBin("ibin") + Add(IntBin("ibin"), IntBin("ibin")) + 5, 20),
            Eq(FloatBin("fbin") + Add(FloatBin("fbin"), FloatBin("fbin")) + 5.0, 20.0),
        ],
    )
    def test_add_overloaded_pos(self, expression):
        """
        Test arithmetic Add expression with correct parameters.
        """
        self.verify_expression(expression.compile(), self.rec)

    @pytest.mark.parametrize(
        "expression, expected",
        [
            (Eq(IntBin("ibin") + Add(IntBin("ibin"), IntBin("ibin")) + 5, 15), e.FilteredOut),
            (Eq(IntBin("ibin") + Add(FloatBin("fbin"), FloatBin("fbin")) + 5.0, 20.0), e.InvalidRequest),
        ],
    )
    def test_add_overloaded_neg(self, expression, expected):
        """
        Test arithmetic Add expression with incorrect parameters.
        """
        self.verify_expression_neg(expression.compile(), expected)

    @pytest.mark.parametrize(
        "bin, val, check",
        [
            (IntBin("ibin"), [5], 0),
            (IntBin("ibin"), [5, 5], -5),
            (IntBin("ibin"), [IntBin("ibin")], 0),
            (FloatBin("fbin"), [3.0, 1.0], 1.0),
        ],
    )
    def test_sub_pos(self, bin, val, check):
        """
        Test arithmetic Sub expression with correct parameters.
        """
        expr = Eq(Sub(bin, *val), check).compile()

        self.verify_expression(expr, self.rec)

    @pytest.mark.parametrize(
        "bin, val, check, expected",
        [
            (IntBin("ibin"), [5], 25, e.FilteredOut),
            (IntBin("ibin"), [5.0], 0, e.InvalidRequest),
            (FloatBin("fbin"), [3], 2.0, e.InvalidRequest),
            (FloatBin("fbin"), ["bad_arg"], 8.0, e.InvalidRequest),
        ],
    )
    def test_sub_neg(self, bin, val, check, expected):
        """
        Test arithmetic Sub expression expecting failure.
        """
        expr = Eq(Sub(bin, *val), check).compile()

        self.verify_expression_neg(expr, expected)

    @pytest.mark.parametrize(
        "expression",
        [
            Eq(IntBin("ibin") - Sub(IntBin("ibin"), IntBin("ibin")) - 5, -10),
            Eq(FloatBin("fbin") - Sub(FloatBin("fbin"), FloatBin("fbin")) - 5.0, -10.0),
        ],
    )
    def test_sub_overloaded_pos(self, expression):
        """
        Test arithmetic Sub expression with correct parameters.
        """
        self.verify_expression(expression.compile(), self.rec)

    @pytest.mark.parametrize(
        "expression, expected",
        [
            (Eq(IntBin("ibin") - Sub(IntBin("ibin"), IntBin("ibin")) - 5, -5), e.FilteredOut),
            (Eq(FloatBin("fbin") - Sub(IntBin("ibin"), FloatBin("fbin")) - 5.0, -10.0), e.InvalidRequest),
        ],
    )
    def test_sub_overloaded_neg(self, expression, expected):
        """
        Test arithmetic Sub expression with incorrect parameters.
        """
        self.verify_expression_neg(expression.compile(), expected)

    @pytest.mark.parametrize(
        "bin, val, check",
        [
            (IntBin("ibin"), [5], 25),
            (IntBin("ibin"), [5, 2], 50),
            (IntBin("ibin"), [IntBin("ibin")], 25),
            (FloatBin("fbin"), [3.0], 15.0),
        ],
    )
    def test_mul_pos(self, bin, val, check):
        """
        Test arithmetic Mul expression with correct parameters.
        """
        expr = Eq(Mul(bin, *val), check).compile()

        self.verify_expression(expr, self.rec)

    @pytest.mark.parametrize(
        "bin, val, check, expected",
        [
            (IntBin("ibin"), [5], 5, e.FilteredOut),
            (IntBin("ibin"), [5.0], 25, e.InvalidRequest),
            (FloatBin("fbin"), [3], 15.0, e.InvalidRequest),
            (FloatBin("fbin"), ["bad_arg"], 8.0, e.InvalidRequest),
        ],
    )
    def test_mul_neg(self, bin, val, check, expected):
        """
        Test arithmetic Mul expression expecting failure.
        """
        expr = Eq(Mul(bin, *val), check).compile()

        self.verify_expression_neg(expr, expected)

    @pytest.mark.parametrize(
        "expression",
        [
            Eq(IntBin("ibin") * Mul(IntBin("ibin"), IntBin("ibin")) * IntBin("ibin"), 625),
            Eq(FloatBin("fbin") * Mul(FloatBin("fbin"), FloatBin("fbin")) * 5.0, 625.0),
        ],
    )
    def test_mul_overloaded_pos(self, expression):
        """
        Test arithmetic Mul expression with correct parameters.
        """
        self.verify_expression(expression.compile(), self.rec)

    @pytest.mark.parametrize(
        "expression, expected",
        [
            (Eq(IntBin("ibin") * Mul(IntBin("ibin"), IntBin("ibin")) * IntBin("ibin"), 5), e.FilteredOut),
            (Eq(IntBin("ibin") * Mul(FloatBin("fbin"), FloatBin("fbin")) * 5.0, 5.0), e.InvalidRequest),
        ],
    )
    def test_mul_overloaded_neg(self, expression, expected):
        """
        Test arithmetic Mul expression with incorrect parameters.
        """
        self.verify_expression_neg(expression.compile(), expected)

    @pytest.mark.parametrize(
        "bin, val, check",
        [
            (IntBin("ibin"), [5], 1),
            (FloatBin("fbin"), [5.0, 2.0], 0.5),
            (IntBin("ibin"), [IntBin("ibin")], 1),
            (FloatBin("fbin"), [3.0], 1.6666666666666667),
        ],
    )
    def test_div_pos(self, bin, val, check):
        """
        Test arithmetic Div expression with correct parameters.
        """
        expr = Eq(Div(bin, *val), check).compile()

        self.verify_expression(expr, self.rec)

    @pytest.mark.parametrize(
        "bin, val, check, expected",
        [
            (IntBin("ibin"), [5], 25, e.FilteredOut),
            (IntBin("ibin"), [5.0], 25, e.InvalidRequest),
            (FloatBin("fbin"), [3], 15.0, e.InvalidRequest),
            (FloatBin("fbin"), ["bad_arg"], 8.0, e.InvalidRequest),
        ],
    )
    def test_div_neg(self, bin, val, check, expected):
        """
        Test arithmetic Div expression expecting failure.
        """
        expr = Eq(Div(bin, *val), check).compile()

        self.verify_expression_neg(expr, expected)

    @pytest.mark.parametrize(
        "expression",
        [
            Eq(IntBin("ibin") * 100 / Div(IntBin("ibin"), IntBin("ibin")) / IntBin("ibin"), 4),
            Eq(FloatBin("fbin") * 100.0 / Div(FloatBin("fbin"), FloatBin("fbin")) / 5.0, 4.0),
        ],
    )
    def test_div_overloaded_pos(self, expression):
        """
        Test arithmetic Div expression with correct parameters.
        """
        self.verify_expression(expression.compile(), self.rec)

    @pytest.mark.parametrize(
        "expression, expected",
        [  # Note that the parens in Div() do not overload precedence with /.
            (Eq(IntBin("ibin") * 100 / Div(IntBin("ibin"), IntBin("ibin")) / IntBin("ibin"), 3), e.FilteredOut),
            (Eq(IntBin("ibin") * 100.0 / Div(FloatBin("fbin"), FloatBin("fbin")) / 5.0, 4.0), e.InvalidRequest),
        ],
    )
    def test_div_overloaded_neg(self, expression, expected):
        """
        Test arithmetic Div expression with incorrect parameters.
        """
        self.verify_expression_neg(expression.compile(), expected)

    @pytest.mark.parametrize(
        "expression",
        [Eq(FloatBin("fbin") * 5.1 // 5.0, 5.0), Eq(FloatBin("fbin") // FloatBin("fbin") // FloatBin("fbin"), 0.0)],
    )
    def test_floor_div_overloaded_pos(self, expression):
        """
        Test arithmetic // expression with correct parameters.
        """
        self.verify_expression(expression.compile(), self.rec)

    @pytest.mark.parametrize(
        "expression, expected",
        [
            (Eq(FloatBin("fbin") * 5.1 // 5.0, 5.1), e.FilteredOut),
            (Eq(IntBin("ibin") * 5.1 // 5.0, 5.0), e.InvalidRequest),
            (Eq(IntBin("ibin") * 5 // 5, 5), e.InvalidRequest),
        ],
    )
    def test_floor_div_overloaded_neg(self, expression, expected):
        """
        Test arithmetic // expression with incorrect parameters.
        """
        self.verify_expression_neg(expression.compile(), expected)

    @pytest.mark.parametrize(
        "bin, val, check", [(FloatBin("fbin"), [2.0], 25.0), (FloatBin("fbin"), [FloatBin("fbin")], 3125.0)]
    )
    def test_pow_pos(self, bin, val, check):
        """
        Test arithmetic Pow expression with correct parameters.
        """
        expr = Eq(Pow(bin, *val), check).compile()

        self.verify_expression(expr, self.rec)

    @pytest.mark.parametrize(
        "bin, val, check, expected",
        [
            (IntBin("ibin"), [2], 25, e.InvalidRequest),
            (IntBin("ibin"), [IntBin("ibin")], 1, e.InvalidRequest),
            (FloatBin("fbin"), [2.0], 26.0, e.FilteredOut),
            (IntBin("ibin"), [5.0], 25, e.InvalidRequest),
            (FloatBin("fbin"), [3], 15.0, e.InvalidRequest),
            (FloatBin("fbin"), ["bad_arg"], 8.0, e.InvalidRequest),
        ],
    )
    def test_pow_neg(self, bin, val, check, expected):
        """
        Test arithmetic Pow expression expecting failure.
        """
        expr = Eq(Pow(bin, *val), check).compile()

        self.verify_expression_neg(expr, expected)

    @pytest.mark.parametrize(
        "expression",
        [
            Eq(FloatBin("fbin") ** FloatBin("fbin"), 3125.0),
            Eq(FloatBin("fbin") ** 5.0, 3125.0),
        ],
    )
    def test_pow_overloaded_pos(self, expression):
        """
        Test arithmetic Pow expression with correct parameters.
        """
        self.verify_expression(expression.compile(), self.rec)

    @pytest.mark.parametrize(
        "expression, expected",
        [
            (Eq(FloatBin("fbin") ** FloatBin("fbin"), 50.0), e.FilteredOut),
            (Eq(FloatBin("fbin") ** 5, 3125.0), e.InvalidRequest),
            (Eq(IntBin("ibin") ** 5.0, 3125.0), e.InvalidRequest),
        ],
    )
    def test_pow_overloaded_neg(self, expression, expected):
        """
        Test arithmetic Pow expression with incorrect parameters.
        """
        self.verify_expression_neg(expression.compile(), expected)

    @pytest.mark.parametrize(
        "bin, val, check",
        [(Pow(FloatBin("fbin"), 4.0), [5.0], 4.0), (Pow(FloatBin("fbin"), 10.0), [FloatBin("fbin")], 10.0)],
    )
    def test_log_pos(self, bin, val, check):
        """
        Test arithmetic Log expression with correct parameters.
        """
        expr = Eq(Log(bin, *val), check).compile()

        self.verify_expression(expr, self.rec)

    @pytest.mark.parametrize(
        "bin, val, check, expected",
        [
            (IntBin("ibin"), [2], 25, e.InvalidRequest),
            (IntBin("ibin"), [IntBin("ibin")], 1, e.InvalidRequest),
            (FloatBin("fbin"), [2.0], 26.0, e.FilteredOut),
            (IntBin("ibin"), [5.0], 25, e.InvalidRequest),
            (FloatBin("fbin"), [3], 15.0, e.InvalidRequest),
            (FloatBin("fbin"), ["bad_arg"], 8.0, e.InvalidRequest),
        ],
    )
    def test_log_neg(self, bin, val, check, expected):
        """
        Test arithmetic Log expression expecting failure.
        """
        expr = Eq(Log(bin, *val), check).compile()

        self.verify_expression_neg(expr, expected)

    @pytest.mark.parametrize(
        "bin, val, check",
        [
            (IntBin("ibin"), [2], 1),
            (IntBin("ibin"), [IntBin("ibin")], 0),
        ],
    )
    def test_mod_pos(self, bin, val, check):
        """
        Test arithmetic Mod expression with correct parameters.
        """
        expr = Eq(Mod(bin, *val), check).compile()

        self.verify_expression(expr, self.rec)

    @pytest.mark.parametrize(
        "bin, val, check, expected",
        [
            (FloatBin("fbin"), [4.0], [5.0], e.InvalidRequest),
            (FloatBin("fbin"), [10.0], [FloatBin("fbin")], e.InvalidRequest),
            (IntBin("ibin"), [2], 0, e.FilteredOut),
            (IntBin("ibin"), [5.0], 25, e.InvalidRequest),
            (FloatBin("fbin"), [3], 15.0, e.InvalidRequest),
            (FloatBin("fbin"), ["bad_arg"], 8.0, e.InvalidRequest),
        ],
    )
    def test_mod_neg(self, bin, val, check, expected):
        """
        Test arithmetic Mod expression expecting failure.
        """
        expr = Eq(Mod(bin, *val), check).compile()

        self.verify_expression_neg(expr, expected)

    @pytest.mark.parametrize(
        "expression",
        [
            Eq(IntBin("ibin") % IntBin("ibin"), 0),
            Eq(IntBin("ibin") % 2, 1),
        ],
    )
    def test_mod_overloaded_pos(self, expression):
        """
        Test arithmetic Mod expression with correct parameters.
        """
        self.verify_expression(expression.compile(), self.rec)

    @pytest.mark.parametrize(
        "expression, expected",
        [
            (Eq(IntBin("ibin") % IntBin("ibin"), 1), e.FilteredOut),
            (Eq(FloatBin("fbin") % 5, 0.0), e.InvalidRequest),
            (Eq(IntBin("ibin") % 5.0, 0.0), e.InvalidRequest),
        ],
    )
    def test_mod_overloaded_neg(self, expression, expected):
        """
        Test arithmetic Mod expression with incorrect parameters.
        """
        self.verify_expression_neg(expression.compile(), expected)

    @pytest.mark.parametrize(
        "expression",
        [
            Eq(abs(Sub(FloatBin("fbin"))), 5.0),
            Eq(abs(FloatBin("fbin")), 5.0),
            Eq(abs(-5.0), 5.0),
            Eq(abs(Sub(IntBin("ibin"))), 5),
            Eq(abs(IntBin("ibin")), 5),
            Eq(abs(-5), 5),
            # using the expr object
            Eq(Abs(Sub(FloatBin("fbin"))), 5.0),
            Eq(Abs(FloatBin("fbin")), 5.0),
            Eq(Abs(-5.0), 5.0),
            Eq(Abs(Sub(IntBin("ibin"))), 5),
            Eq(Abs(IntBin("ibin")), 5),
            Eq(Abs(-5), 5),
        ],
    )
    def test_abs_pos(self, expression):
        """
        Test arithmetic Abs expression with correct parameters.
        """
        self.verify_expression(expression.compile(), self.rec)

    @pytest.mark.parametrize(
        "expression, expected",
        [
            (Eq(Abs(Sub(FloatBin("fbin"))), -5.0), e.FilteredOut),
            (Eq(Abs(Sub(StrBin("bad_bin"))), -5.0), e.InvalidRequest),
            # using the expr object
            (Eq(Abs(Sub(FloatBin("fbin"))), -5.0), e.FilteredOut),
            (Eq(Abs(Sub(StrBin("bad_bin"))), -5.0), e.InvalidRequest),
        ],
    )
    def test_abs_neg(self, expression, expected):
        """
        Test arithmetic Abs expression with incorrect parameters.
        """
        self.verify_expression_neg(expression.compile(), expected)

    @pytest.mark.parametrize("val, check", [(8.953, 8.0), (Add(FloatBin("fbin"), 4.5), 9.0)])
    def test_floor_pos(self, val, check):
        """
        Test arithmetic Floor expression with correct parameters.
        """
        expr = Eq(Floor(val), check).compile()

        self.verify_expression(expr, self.rec)

    @pytest.mark.parametrize(
        "val, check, expected",
        [
            (IntBin("ibin"), 25, e.InvalidRequest),
            (1, 1, e.InvalidRequest),
            (FloatBin("fbin"), 26.0, e.FilteredOut),
            ("bad_arg", 8.0, e.InvalidRequest),
        ],
    )
    def test_floor_neg(self, val, check, expected):
        """
        Test arithmetic Floor expression expecting failure.
        """
        expr = Eq(Floor(val), check).compile()

        self.verify_expression_neg(expr, expected)

    @pytest.mark.parametrize("expression", [Eq(math.floor(FloatBin("fbin") + 1.6), 6.0)])
    def test_floor_overloaded_pos(self, expression):
        """
        Test arithmetic Floor expression with correct parameters.
        """
        self.verify_expression(expression.compile(), self.rec)

    @pytest.mark.parametrize(
        "expression, expected",
        [
            (Eq(math.floor(FloatBin("fbin") + 1.6), 7.0), e.FilteredOut),
            (Eq(math.floor(IntBin("ibin")), 5), e.InvalidRequest),
        ],
    )
    def test_floor_overloaded_neg(self, expression, expected):
        """
        Test arithmetic Floor expression with incorrect parameters.
        """
        self.verify_expression_neg(expression.compile(), expected)

    @pytest.mark.parametrize("val, check", [(8.953, 9.0), (Add(FloatBin("fbin"), 4.5), 10.0)])
    def test_ceil_pos(self, val, check):
        """
        Test arithmetic Ceil expression with correct parameters.
        """
        expr = Eq(Ceil(val), check).compile()

        self.verify_expression(expr, self.rec)

    @pytest.mark.parametrize(
        "val, check, expected",
        [
            (IntBin("ibin"), 25, e.InvalidRequest),
            (1, 1, e.InvalidRequest),
            (FloatBin("fbin"), 26.0, e.FilteredOut),
            ("bad_arg", 8.0, e.InvalidRequest),
        ],
    )
    def test_ceil_neg(self, val, check, expected):
        """
        Test arithmetic Ceil expression expecting failure.
        """
        expr = Eq(Ceil(val), check).compile()

        self.verify_expression_neg(expr, expected)

    @pytest.mark.parametrize("expression", [Eq(math.ceil(FloatBin("fbin") + 1.6), 7.0)])
    def test_ceil_overloaded_pos(self, expression):
        """
        Test arithmetic Ceil expression with correct parameters.
        """
        self.verify_expression(expression.compile(), self.rec)

    @pytest.mark.parametrize(
        "expression, expected",
        [
            (Eq(math.ceil(FloatBin("fbin") + 1.6), 6.0), e.FilteredOut),
            (Eq(math.ceil(IntBin("ibin")), 5), e.InvalidRequest),
        ],
    )
    def test_ceil_overloaded_neg(self, expression, expected):
        """
        Test arithmetic Ceil expression with incorrect parameters.
        """
        self.verify_expression_neg(expression.compile(), expected)

    @pytest.mark.parametrize("val, check", [(8.953, 8), (FloatBin("fbin"), 5)])
    def test_toint_pos(self, val, check):
        """
        Test arithmetic ToInt expression with correct parameters.
        """
        expr = Eq(ToInt(val), check).compile()

        self.verify_expression(expr, self.rec)

    @pytest.mark.parametrize(
        "val, check, expected",
        [
            (IntBin("ibin"), 5, e.InvalidRequest),
            (1, 1, e.InvalidRequest),
            (FloatBin("fbin"), 26, e.FilteredOut),
            ("bad_arg", 8.0, e.InvalidRequest),
        ],
    )
    def test_toint_neg(self, val, check, expected):
        """
        Test arithmetic ToInt expression expecting failure.
        """
        expr = Eq(ToInt(val), check).compile()

        self.verify_expression_neg(expr, expected)

    @pytest.mark.parametrize("val, check", [(8, 8.0), (IntBin("ibin"), 5.0)])
    def test_tofloat_pos(self, val, check):
        """
        Test arithmetic ToFloat expression with correct parameters.
        """
        expr = Eq(ToFloat(val), check).compile()

        self.verify_expression(expr, self.rec)

    @pytest.mark.parametrize(
        "val, check, expected",
        [
            (FloatBin("fbin"), 5.0, e.InvalidRequest),
            (1.0, 1.0, e.InvalidRequest),
            (IntBin("ibin"), 6.0, e.FilteredOut),
            ("bad_arg", 8.0, e.InvalidRequest),
        ],
    )
    def test_tofloat_neg(self, val, check, expected):
        """
        Test arithmetic ToFloat expression expecting failure.
        """
        expr = Eq(ToFloat(val), check).compile()

        self.verify_expression_neg(expr, expected)

    @pytest.mark.parametrize(
        "bin, val, check",
        [
            (IntBin("ibin"), [5, 25, 4, 35, 64, 1, 23, 2, 2, 2], 1),
            (IntBin("ibin"), [Add(IntBin("ibin"), 5)], 5),
            (IntBin("ibin"), [IntBin("ibin")], 5),
            (FloatBin("fbin"), [6.0, 20.0], 5.0),
        ],
    )
    def test_min_pos(self, bin, val, check):
        """
        Test arithmetic Min expression with correct parameters.
        """
        expr = Eq(Min(bin, *val), check).compile()

        self.verify_expression(expr, self.rec)

    @pytest.mark.parametrize(
        "bin, val, check, expected",
        [
            (IntBin("ibin"), [5], 25, e.FilteredOut),
            (IntBin("ibin"), [5.0], 0, e.InvalidRequest),
            (FloatBin("fbin"), [3], 2.0, e.InvalidRequest),
            (FloatBin("fbin"), ["bad_arg"], 8.0, e.InvalidRequest),
        ],
    )
    def test_min_neg(self, bin, val, check, expected):
        """
        Test arithmetic Min expression expecting failure.
        """
        expr = Eq(Min(bin, *val), check).compile()

        self.verify_expression_neg(expr, expected)

    @pytest.mark.parametrize(
        "bin, val, check",
        [
            (IntBin("ibin"), [5, 25, 4, 35, 64, 1, 23, 2, 2, 2], 64),
            (IntBin("ibin"), [Add(IntBin("ibin"), 5)], 10),
            (IntBin("ibin"), [IntBin("ibin")], 5),
            (FloatBin("fbin"), [6.0, 20.0], 20.0),
        ],
    )
    def test_max_pos(self, bin, val, check):
        """
        Test arithmetic Max expression with correct parameters.
        """
        expr = Eq(Max(bin, *val), check).compile()

        self.verify_expression(expr, self.rec)

    @pytest.mark.parametrize(
        "bin, val, check, expected",
        [
            (IntBin("ibin"), [6], 5, e.FilteredOut),
            (IntBin("ibin"), [5.0], 0, e.InvalidRequest),
            (FloatBin("fbin"), [3], 2.0, e.InvalidRequest),
            (FloatBin("fbin"), ["bad_arg"], 8.0, e.InvalidRequest),
        ],
    )
    def test_max_neg(self, bin, val, check, expected):
        """
        Test arithmetic Max expression expecting failure.
        """
        expr = Eq(Max(bin, *val), check).compile()

        self.verify_expression_neg(expr, expected)



# -*- coding: utf-8 -*-
import pytest
from aerospike import exception as e
from aerospike_helpers.operations import hll_operations
from math import sqrt


import aerospike


class TestHLL(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        self.test_keys = []
        for x in range(5):

            key = ("test", "demo", x)
            record = {"label": x}

            as_connection.put(key, record)

            ops = [
                hll_operations.hll_init("hll_bine", 10),
                hll_operations.hll_add("hll_bin", ["key%s" % str(i) for i in range(x + 1)], 8),
                hll_operations.hll_add("mh_bin", ["key%s" % str(i) for i in range(x + 1)], 6, 12),
                hll_operations.hll_add("hll_binl", ["key%s" % str(i) for i in range(100)], 8),
                hll_operations.hll_add("hll_binlh", ["key%s" % str(i) for i in range(50)], 8),
                hll_operations.hll_add("hll_bin_big", ["key%s" % str(i) for i in range(1000)], 10),
                hll_operations.hll_add("hll_binu", ["key6", "key7", "key8", "key9", "key10"], 10),
            ]

            self.test_keys.append(key)
            _, _, _ = as_connection.operate(key, ops)

        def teardown():
            """
            Teardown method.
            """
            for key in self.test_keys:
                try:
                    as_connection.remove(key)
                except e.RecordNotFound:
                    pass

        request.addfinalizer(teardown)

    def relative_count_error(self, n_index_bits):
        return 1.04 / sqrt(2**n_index_bits)

    def relative_intersect_error(self, n_index_bits, bin_counts, bin_intersect_count):
        sigma = 1.04 / sqrt(2**n_index_bits)
        rel_err_sum = 0
        for count in bin_counts:
            rel_err_sum += (sigma * count) ** 2
        rel_err_sum += sigma * (bin_intersect_count**2)

        return sqrt(rel_err_sum)

    def assert_within_error_bounds(self, vut, actual, error_percent):
        rel_error = vut * error_percent
        assert vut >= actual - rel_error and vut <= actual + rel_error

    @pytest.mark.parametrize(
        "policy, expected_result",
        [(None, 3), ({"flags": aerospike.HLL_WRITE_CREATE_ONLY}, 3), ({"flags": aerospike.HLL_WRITE_NO_FAIL}, 3)],
    )
    def test_pos_hll_add(self, policy, expected_result):
        """
        Invoke hll_add() creating a new HLL.
        """
        ops = [hll_operations.hll_add("new_bin", ["key1", "key2", "key3"], index_bit_count=8, policy=policy)]

        _, _, res = self.as_connection.operate(self.test_keys[0], ops)

        assert res["new_bin"] == expected_result

    @pytest.mark.parametrize(
        "policy, expected_result", [({"flags": aerospike.HLL_WRITE_UPDATE_ONLY}, e.InvalidRequest)]
    )
    def test_neg_hll_add(self, policy, expected_result):
        """
        Invoke hll_add() creating with expected failures.
        """
        ops = [hll_operations.hll_add("new_bin", ["key1", "key2", "key3"], index_bit_count=8, policy=policy)]

        with pytest.raises(expected_result):
            self.as_connection.operate(self.test_keys[0], ops)

    def test_pos_hll_add_mh(self):
        """
        Invoke hll_add() creating a new min hash HLL.
        """
        ops = [hll_operations.hll_add("new_bin", ["key1", "key2", "key3"], 6, 8)]

        _, _, res = self.as_connection.operate(self.test_keys[0], ops)

        assert res["new_bin"] == 3

    @pytest.mark.parametrize(
        "policy, expected_result", [({"flags": aerospike.HLL_WRITE_CREATE_ONLY}, e.BinExistsError)]
    )
    def test_neg_hll_add_mh(self, policy, expected_result):
        """
        Invoke hll_add() failing to creating a new min hash HLL.
        """
        ops = [hll_operations.hll_add("hll_binl", ["key101", "key102", "key103"], 6, 8, policy)]

        with pytest.raises(expected_result):
            self.as_connection.operate(self.test_keys[0], ops)

    def test_pos_hll_get_count(self):
        """
        Invoke hll_get_count() to check an HLL's count.
        """
        ops = [hll_operations.hll_get_count("hll_bin_big")]
        actual_count = 1000
        rel_error = self.relative_count_error(10)

        _, _, res = self.as_connection.operate(self.test_keys[2], ops)
        self.assert_within_error_bounds(res["hll_bin_big"], actual_count, rel_error)

    def test_neg_hll_get_count(self):
        """
        Invoke hll_get_count() with a non existent bin.
        """
        ops = [hll_operations.hll_get_count("fake_hll_bin")]

        _, _, res = self.as_connection.operate(self.test_keys[2], ops)
        assert res["fake_hll_bin"] is None

    def test_pos_hll_describe(self):
        """
        Invoke hll_describe() and check index and min hash bits.
        """
        ops = [hll_operations.hll_describe("mh_bin")]

        _, _, res = self.as_connection.operate(self.test_keys[0], ops)
        assert res["mh_bin"] == [6, 12]

    def test_neg_hll_describe(self):
        """
        Invoke hll_describe() with a non existent hll bin.
        """
        ops = [hll_operations.hll_describe("fake_hll_bin")]

        _, _, res = self.as_connection.operate(self.test_keys[0], ops)
        assert res["fake_hll_bin"] is None

    def test_pos_hll_fold(self):
        """
        Invoke hll_fold().
        """
        ops = [hll_operations.hll_fold("hll_bin", 6), hll_operations.hll_describe("hll_bin")]

        _, _, res = self.as_connection.operate(self.test_keys[2], ops)
        assert res["hll_bin"] == [6, 0]

    @pytest.mark.parametrize(
        "index_bits, bin, expected_result", [(2, "hll_bin", e.InvalidRequest), (4, "mh_bin", e.OpNotApplicable)]
    )
    def test_neg_hll_fold(self, index_bits, bin, expected_result):
        """
        Invoke hll_fold() expecting errors.
        """
        ops = [hll_operations.hll_fold(bin, index_bits), hll_operations.hll_describe(bin)]

        with pytest.raises(expected_result):
            self.as_connection.operate(self.test_keys[2], ops)

    def test_pos_hll_get_intersect_count(self):
        """
        Invoke hll_get_intersect_count().
        """

        records = [record[2]["hll_binu"] for record in self.as_connection.get_many(self.test_keys[:1])]
        ops = [hll_operations.hll_get_intersect_count("hll_binl", records)]
        rel_error = self.relative_intersect_error(8, [100, 5], 5)
        actual_intersect = 5

        _, _, res = self.as_connection.operate(self.test_keys[4], ops)
        assert res["hll_binl"] >= actual_intersect - rel_error and res["hll_binl"] <= actual_intersect + rel_error

    def test_pos_hll_get_intersect_count_mh(self):
        """
        Invoke hll_get_intersect_count() on min hash bins.
        """
        records = [record[2]["mh_bin"] for record in self.as_connection.get_many(self.test_keys[:4])]

        ops = [hll_operations.hll_get_intersect_count("mh_bin", records)]

        _, _, res = self.as_connection.operate(self.test_keys[4], ops)
        assert res["mh_bin"] == 1

    @pytest.mark.parametrize("hll_bins, bin, expected_result", [(5, "hll_binl", e.InvalidRequest)])
    def test_neg_hll_get_intersect_count(self, hll_bins, bin, expected_result):
        """
        Invoke hll_get_intersect_count() with expected failures.
        """

        records = [record[2][bin] for record in self.as_connection.get_many(self.test_keys[: hll_bins - 1])]

        ops = [hll_operations.hll_get_intersect_count(bin, records)]

        with pytest.raises(expected_result):
            self.as_connection.operate(self.test_keys[4], ops)

    def test_pos_hll_get_similarity(self):
        """
        Invoke hll_get_similarity().
        """

        hll = self.as_connection.get(self.test_keys[0])[2]["hll_binl"]

        ops = [hll_operations.hll_get_similarity("hll_binlh", [hll])]
        rel_error = self.relative_intersect_error(8, [100, 50], 50)
        actual_similarity = 0.50

        _, _, res = self.as_connection.operate(self.test_keys[4], ops)
        assert res["hll_binlh"] >= actual_similarity - rel_error and res["hll_binlh"] <= actual_similarity + rel_error

    @pytest.mark.parametrize("hll_bins, bin, expected_result", [(5, "hll_binl", e.InvalidRequest)])
    def test_neg_hll_get_similarity(self, hll_bins, bin, expected_result):
        """
        Invoke hll_get_similarity() with expected errors.
        """

        records = [record[2][bin] for record in self.as_connection.get_many(self.test_keys[: hll_bins - 1])]

        ops = [hll_operations.hll_get_similarity(bin, records)]

        with pytest.raises(expected_result):
            self.as_connection.operate(self.test_keys[4], ops)

    def test_pos_hll_get_union(self):
        """
        Invoke hll_get_union().
        """

        records = [record[2]["hll_binu"] for record in self.as_connection.get_many(self.test_keys)]

        ops = [hll_operations.hll_get_union("hll_bin", [records[4]])]

        _, _, union_hll = self.as_connection.operate(self.test_keys[4], ops)

        ops = [
            hll_operations.hll_get_intersect_count("hll_bin", [union_hll["hll_bin"]]),
            hll_operations.hll_get_intersect_count("hll_binu", [union_hll["hll_bin"]]),
        ]

        _, _, res = self.as_connection.operate(self.test_keys[4], ops)

        assert res["hll_bin"] == 5
        assert res["hll_binu"] == 5

    def test_neg_hll_get_union(self):
        """
        Invoke hll_get_union() with expected errors.
        """

        [record[2]["hll_binu"] for record in self.as_connection.get_many(self.test_keys)]

        ops = [hll_operations.hll_get_union("hll_bin", [])]

        with pytest.raises(e.InvalidRequest):
            _, _, union_hll = self.as_connection.operate(self.test_keys[4], ops)

    def test_pos_hll_get_union_count(self):
        """
        Invoke hll_get_union_count().
        """

        records = [record[2]["hll_bin"] for record in self.as_connection.get_many(self.test_keys)]

        ops = [hll_operations.hll_get_union_count("hll_binu", records)]

        _, _, res = self.as_connection.operate(self.test_keys[0], ops)

        assert res["hll_binu"] == 10

    def test_neg_hll_get_union_count(self):
        """
        Invoke hll_get_union_count() with errors.
        """

        records = []

        ops = [hll_operations.hll_get_union_count("hll_binu", records)]

        with pytest.raises(e.InvalidRequest):
            self.as_connection.operate(self.test_keys[0], ops)

    @pytest.mark.parametrize(
        "index_bits, mh_bits, bin, policy",
        [
            (4, 4, "new_mhbin", None),
            (4, 51, "new_mhbin", None),
            (9, 51, "new_mhbin", None),
            (6, 8, "new_mhbin", {"flags": aerospike.HLL_WRITE_CREATE_ONLY}),
            (6, 8, "new_mhbin", {"flags": aerospike.HLL_WRITE_DEFAULT}),
            (6, 8, "mh_bin", {"flags": aerospike.HLL_WRITE_UPDATE_ONLY}),
        ],
    )
    def test_pos_hll_init_mh(self, index_bits, mh_bits, bin, policy):
        """
        Invoke hll_init_mh() creating a new min hash HLL.
        """
        ops = [hll_operations.hll_init(bin, index_bits, mh_bits, policy)]

        _, _, _ = self.as_connection.operate(self.test_keys[0], ops)

        record = self.as_connection.get(self.test_keys[0])
        assert record[2][bin]

    @pytest.mark.parametrize(
        "index_bits, mh_bits, bin, policy, expected_result",
        [
            (0, 0, "new_mhbin", None, e.InvalidRequest),
            ("bad_ind_bits", 4, "new_mhbin", None, e.ParamError),
            (4, "bad_mh_bits", "new_mhbin", None, e.ParamError),
            (4, 4, "new_mhbin", "bad_policy", e.ParamError),
            (4, 4, ["bad_bin"], None, e.ParamError),
            (4, 52, "new_mhbin", None, e.InvalidRequest),
            (4, 59, "new_mhbin", None, e.InvalidRequest),
            (17, 48, "new_mhbin", None, e.InvalidRequest),
            (20, 8, "new_mhbin", None, e.InvalidRequest),
            (6, 8, "mh_bin", {"flags": aerospike.HLL_WRITE_CREATE_ONLY}, e.BinExistsError),
            (6, 8, "new_mhbin", {"flags": aerospike.HLL_WRITE_UPDATE_ONLY}, e.BinNotFound),
        ],
    )
    def test_neg_hll_init_mh(self, index_bits, mh_bits, bin, policy, expected_result):
        """
        Invoke hll_init_mh() expecting failures.
        """
        ops = [hll_operations.hll_init(bin, index_bits, mh_bits, policy)]

        with pytest.raises(expected_result):
            self.as_connection.operate(self.test_keys[0], ops)

    def test_pos_hll_init(self):
        """
        Invoke hll_init() creating a new HLL.
        """
        ops = [hll_operations.hll_init("new_hll", 6)]

        _, _, _ = self.as_connection.operate(self.test_keys[0], ops)

        record = self.as_connection.get(self.test_keys[0])
        assert record[2]["new_hll"]

    @pytest.mark.parametrize(
        "index_bits, bin, policy, expected_result",
        [
            (3, "new_hllbin", None, e.InvalidRequest),
            (20, "new_hllbin", None, e.InvalidRequest),
            (6, "hll_bin", {"flags": aerospike.HLL_WRITE_CREATE_ONLY}, e.BinExistsError),
            (6, "new_hllbin", {"flags": aerospike.HLL_WRITE_UPDATE_ONLY}, e.BinNotFound),
        ],
    )
    def test_neg_hll_init(self, index_bits, bin, policy, expected_result):
        """
        Invoke hll_init() expecting failures.
        """
        ops = [hll_operations.hll_init(bin, index_bits, policy=policy)]

        with pytest.raises(expected_result):
            self.as_connection.operate(self.test_keys[0], ops)

    def test_neg_hll_refresh_count(self):
        """
        Invoke hll_refresh_count() expecting failures.
        """
        ops = [hll_operations.hll_refresh_count("new_hll")]

        with pytest.raises(e.BinNotFound):
            self.as_connection.operate(self.test_keys[0], ops)

    def test_pos_hll_refresh_count(self):
        """
        Invoke hll_refresh_count().
        """
        ops = [hll_operations.hll_refresh_count("hll_binl")]

        _, _, res = self.as_connection.operate(self.test_keys[0], ops)
        assert res["hll_binl"] == 97

    @pytest.mark.parametrize(
        "bin, policy, expected_result",
        [
            ("hll_bine", {}, e.OpNotApplicable),
            ("hll_binl", {"flags": aerospike.HLL_WRITE_CREATE_ONLY}, e.BinExistsError),
            ("new_hll", {"flags": aerospike.HLL_WRITE_UPDATE_ONLY}, e.BinNotFound),
        ],
    )
    def test_neg_hll_set_union(self, bin, policy, expected_result):
        """
        Invoke hll_set_union() expecting failures.
        """

        records = [record[2]["hll_binl"] for record in self.as_connection.get_many(self.test_keys)]

        ops = [hll_operations.hll_set_union(bin, records, policy)]

        with pytest.raises(expected_result):
            self.as_connection.operate(self.test_keys[4], ops)

    @pytest.mark.parametrize(
        "bin, ubin, policy, expected_res",
        [("hll_bin", "hll_bin", {}, 10), ("hll_bine", "hll_binl", {"flags": aerospike.HLL_WRITE_ALLOW_FOLD}, 5)],
    )
    def test_pos_hll_set_union(self, bin, ubin, expected_res, policy):
        """
        Invoke hll_set_union().
        """

        records = [record[2]["hll_binu"] for record in self.as_connection.get_many(self.test_keys)]

        ops = [hll_operations.hll_set_union(bin, [records[4]], policy)]

        _, _, _ = self.as_connection.operate(self.test_keys[4], ops)

        union_hll = self.as_connection.get(self.test_keys[4])[2][ubin]

        ops = [
            hll_operations.hll_get_intersect_count(bin, [union_hll]),
            hll_operations.hll_get_intersect_count("hll_binu", [union_hll]),
        ]

        _, _, res = self.as_connection.operate(self.test_keys[4], ops)

        assert res[bin] == expected_res
        assert res["hll_binu"] == 5

    @pytest.mark.parametrize(
        "bin, policy, expected_result", [("new_hll", {"flags": aerospike.HLL_WRITE_CREATE_ONLY}, e.OpNotApplicable)]
    )
    def test_neg_hll_update(self, bin, policy, expected_result):
        """
        Invoke hll_update() with errors.
        """
        ops = [hll_operations.hll_add(bin, ["key1", "key2", "key3"], policy=policy)]

        with pytest.raises(expected_result):
            self.as_connection.operate(self.test_keys[0], ops, policy)

    def test_pos_hll_update(self):
        """
        Invoke hll_update() creating a new HLL.
        """
        ops = [hll_operations.hll_add("hll_bine", ["key1", "key2", "key3"])]

        _, _, res = self.as_connection.operate(self.test_keys[0], ops)

        assert res["hll_bine"] == 3



# -*- coding: utf-8 -*-
import pytest
import random
from aerospike import exception as e

import aerospike


class TestListInsertItems(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        keys = []
        for i in range(5):
            key = ("test", "demo", i)
            rec = {"name": "name%s" % (str(i)), "age": [i, i + 1], "city": ["Pune", "Dehli"]}
            self.as_connection.put(key, rec)
            keys.append(key)
        key = ("test", "demo", "bytearray_key")
        self.as_connection.put(key, {"bytearray_bin": bytearray("asd;as[d'as;d", "utf-8")})
        keys.append(key)

        def teardown():
            """
            Teardown method.
            """
            for key in keys:
                try:
                    as_connection.remove(key)
                except e.RecordNotFound:
                    pass

        request.addfinalizer(teardown)

    @pytest.mark.parametrize(
        "key, field, index, value, expected",
        [
            (
                ("test", "demo", 1),  # list of integers
                "age",
                0,
                [500, 1500, 3000],
                {"age": [500, 1500, 3000, 1, 2], "name": "name1", "city": ["Pune", "Dehli"]},
            ),
            (
                ("test", "demo", 1),  # list of string
                "city",
                0,
                ["Chennai"],
                {"age": [1, 2], "name": "name1", "city": ["Chennai", "Pune", "Dehli"]},
            ),
            (
                ("test", "demo", 1),  # list of unicode string
                "city",
                3,
                ["Mumbai"],
                {"age": [1, 2], "city": ["Pune", "Dehli", None, "Mumbai"], "name": "name1"},
            ),
            (
                ("test", "demo", 2),  # list of float
                "age",
                7,
                [85.12],
                {"age": [2, 3, None, None, None, None, None, 85.12], "city": ["Pune", "Dehli"], "name": "name2"},
            ),
            (
                ("test", "demo", 3),  # list of map
                "age",
                1,
                [{"k1": 29}],
                {"age": [3, {"k1": 29}, 4], "city": ["Pune", "Dehli"], "name": "name3"},
            ),
            (
                ("test", "demo", 1),  # list of bytearray
                "age",
                2,
                [555, bytearray("asd;as[d'as;d", "utf-8")],
                {"age": [1, 2, 555, bytearray(b"asd;as[d'as;d")], "city": ["Pune", "Dehli"], "name": "name1"},
            ),
        ],
    )
    def test_pos_list_insert_items(self, key, field, index, value, expected):
        """
        Invoke list_insert_items() inserts items
        """
        self.as_connection.list_insert_items(key, field, index, value)

        (key, _, bins) = self.as_connection.get(key)

        assert bins == expected

    def test_pos_list_insert_items_list_with_correct_policy(self):
        """
        Invoke list_insert_items() inserts list with correct policy
        """
        key = ("test", "demo", 2)
        policy = {
            "timeout": 1000,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "commit_level": aerospike.POLICY_COMMIT_LEVEL_MASTER,
        }
        self.as_connection.list_insert_items(key, "age", 5, [45, 50, 80], {}, policy)

        (key, _, bins) = self.as_connection.get(key)

        assert bins == {"age": [2, 3, None, None, None, 45, 50, 80], "city": ["Pune", "Dehli"], "name": "name2"}

    def test_pos_list_insert_items_boolean(self):
        """
        Invoke list_insert_items() insert boolean into the list
        """

        key1 = ("test", "demo", 1)
        self.as_connection.list_insert_items(key1, "age", 1, [False])
        (_, _, b1) = self.as_connection.get(key1)
        key2 = ("test", "demo", 2)
        self.as_connection.list_insert_items(key2, "age", 1, [False, True])
        (_, _, b2) = self.as_connection.get(key2)
        assert b1["age"] == [1, False, 2]
        assert b2["age"] == [2, False, True, 3]

    def test_pos_list_insert_items_with_nonexistent_key(self):
        """
        Invoke list_insert_items() with non-existent key
        """
        charSet = "abcdefghijklmnopqrstuvwxyz1234567890"
        minLength = 5
        maxLength = 30
        length = random.randint(minLength, maxLength)
        key = ("test", "demo", "".join(map(lambda unused: random.choice(charSet), range(length))) + ".com")
        status = self.as_connection.list_insert_items(key, "abc", 2, [122, 878])
        assert status == 0

        (key, _, bins) = self.as_connection.get(key)

        assert status == 0
        assert bins == {"abc": [None, None, 122, 878]}

        self.as_connection.remove(key)

    def test_pos_list_insert_items_with_nonexistent_bin(self):
        """
        Invoke list_insert_items() with non-existent bin
        """
        key = ("test", "demo", 1)
        charSet = "abcdefghijklmnopqrstuvwxyz1234567890"
        minLength = 5
        maxLength = 10
        length = random.randint(minLength, maxLength)
        bin = "".join(map(lambda unused: random.choice(charSet), range(length))) + ".com"

        status = self.as_connection.list_insert_items(key, bin, 1, [585, 789])
        assert status == 0

        (key, _, bins) = self.as_connection.get(key)

        assert status == 0
        assert bins == {"age": [1, 2], "name": "name1", "city": ["Pune", "Dehli"], bin: [None, 585, 789]}

    # Negative Tests
    def test_neg_list_insert_items_with_no_parameters(self):
        """
        Invoke list_insert_items() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.list_insert_items()
        assert "argument 'key' (pos 1)" in str(typeError.value)

    def test_neg_list_insert_items_with_incorrect_policy(self):
        """
        Invoke list_insert_items() with incorrect policy
        """
        key = ("test", "demo", 1)
        policy = {"timeout": 0.5}
        try:
            self.as_connection.list_insert_items(key, "age", 6, ["str"], {}, policy)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "timeout is invalid"

    def test_neg_list_insert_items_with_extra_parameter(self):
        """
        Invoke list_insert_items() with extra parameter.
        """
        key = ("test", "demo", 1)
        policy = {"timeout": 1000}
        with pytest.raises(TypeError) as typeError:
            self.as_connection.list_insert_items(key, "age", 3, [999], {}, policy, "")

        assert "list_insert_items() takes at most 6 arguments (7 given)" in str(typeError.value)

    def test_neg_ist_insert_items_policy_is_string(self):
        """
        Invoke list_insert_items() with policy is string
        """
        key = ("test", "demo", 1)
        try:
            self.as_connection.list_insert_items(key, "age", 1, [85], {}, "")

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "policy must be a dict"

    def test_neg_list_insert_items_key_is_none(self):
        """
        Invoke list_insert_items() with key is none
        """
        try:
            self.as_connection.list_insert_items(None, "age", 1, [45])

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "key is invalid"

    def test_neg_list_insert_items_bin_is_none(self):
        """
        Invoke list_insert_items() with bin is none
        """
        key = ("test", "demo", 1)
        try:
            self.as_connection.list_insert_items(key, None, 2, ["str"])

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Bin name should be of type string"

    def test_neg_list_insert_items_with_items_type_string(self):
        """
        Invoke list_insert_items() insert items is of type string
        """
        key = ("test", "demo", 1)

        try:
            self.as_connection.list_insert_items(key, "age", 6, "abc")

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Items should be of type list"

    def test_neg_list_insert_items_meta_type_integer(self):
        """
        Invoke list_insert_items() with metadata input is of type integer
        """
        key = ("test", "demo", 1)
        try:
            self.as_connection.list_insert_items(key, "contact_no", 0, [85], 888)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Metadata should be of type dictionary"

    def test_neg_list_insert_items_index_type_string(self):
        """
        Invoke list_insert_items() insert with index is of type string
        """
        key = ("test", "demo", 1)

        with pytest.raises(TypeError) as typeError:
            self.as_connection.list_insert_items(key, "age", "Fifth", [False])
        assert "an integer is required" or "cannot be interpreted as an integer" in str(typeError.value)



# -*- coding: utf-8 -*-

import pytest
import time
from .test_base_class import TestBaseClass
from aerospike import exception as e

import aerospike


@pytest.mark.usefixtures("connection_config")
class TestCreateUser(object):

    pytestmark = pytest.mark.skipif(
        not TestBaseClass.auth_in_use(), reason="No user specified, may be not secured cluster."
    )

    def setup_method(self, method):
        """
        Setup method
        """
        config = TestBaseClass.get_connection_config()

        self.client = aerospike.client(config).connect(config["user"], config["password"])

        self.delete_users = []

    def teardown_method(self, method):
        """
        Teardown method
        """

        policy = {}

        for user in self.delete_users:
            try:
                self.client.admin_drop_user(user, policy)
            except Exception:
                pass
        time.sleep(2)
        self.client.close()

    def test_create_user_without_any_parameters(self):

        with pytest.raises(TypeError) as typeError:
            self.client.admin_create_user()

        assert "argument 'user' (pos 1)" in str(typeError.value)

    def test_create_user_with_proper_parameters(self):

        policy = {"timeout": 1000}
        user = "user1-test"
        password = "user1-test"
        roles = ["read", "read-write", "sys-admin"]

        try:
            self.client.admin_drop_user(user, policy)
            time.sleep(2)
        except Exception:
            pass

        status = self.client.admin_create_user(user, password, roles, policy)

        time.sleep(2)

        assert status == 0

        user_details = self.client.admin_query_user(user, policy)

        assert user_details == ["read", "read-write", "sys-admin"]

        self.delete_users.append("user1-test")

    def test_create_user_with_proper_parameters_without_policy(self):

        policy = {"timeout": 1000}
        user = "user1-test"
        password = "user1-test"
        roles = ["read", "read-write", "sys-admin"]

        try:
            self.client.admin_drop_user(user, policy)
            time.sleep(2)
        except Exception:
            pass

        status = self.client.admin_create_user(user, password, roles)

        time.sleep(2)

        assert status == 0

        user_details = self.client.admin_query_user(user, policy)

        assert user_details == ["read", "read-write", "sys-admin"]

        self.delete_users.append("user1-test")

    def test_create_user_with_invalid_timeout_policy_value(self):

        policy = {"timeout": 0.1}
        user = "user3-test"
        password = "user3-test"
        roles = ["sys-admin"]

        try:
            self.client.admin_drop_user(user, policy)
            time.sleep(2)
        except Exception:
            pass

        try:
            self.client.admin_create_user(user, password, roles, policy)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "timeout is invalid"

    def test_create_user_with_proper_timeout_policy_value(self):

        policy = {"timeout": 20}
        user = "user2-test"
        password = "user2-test"
        roles = ["read-write", "sys-admin"]

        try:
            self.client.admin_drop_user(user, policy)
            time.sleep(2)
        except Exception:
            pass

        status = self.client.admin_create_user(user, password, roles, policy)

        time.sleep(2)

        assert status == 0

        user_details = self.client.admin_query_user(user)

        assert user_details == ["read-write", "sys-admin"]

        self.delete_users.append("user2-test")

    def test_create_user_with_none_username(self):

        policy = {"timeout": 20}
        user = None
        password = "user3-test"
        roles = ["sys-admin"]

        try:
            self.client.admin_create_user(user, password, roles, policy)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Username should be a string"

    def test_create_user_with_empty_username(self):

        policy = {}
        user = ""
        password = "user3-test"
        roles = ["read-write"]

        try:
            self.client.admin_create_user(user, password, roles, policy)

        except e.InvalidUser as exception:
            assert exception.code == 60
            assert exception.msg == "AEROSPIKE_INVALID_USER"

    def test_create_user_with_special_characters_in_username(self):

        policy = {}
        user = "!#Q#AEQ@#$%&^*((^&*~~~````"
        password = "uesr4-test"
        roles = ["read-write"]

        try:
            self.client.admin_drop_user(user, policy)
            time.sleep(2)
        except Exception:
            pass

        status = self.client.admin_create_user(user, password, roles, policy)

        assert status == 0

        self.delete_users.append(user)

    def test_create_user_with_none_password(self):

        policy = {}
        user = "uesr4-test"
        password = None
        roles = ["sys-admin"]

        try:
            self.client.admin_create_user(user, password, roles, policy)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Password should be a string"

    def test_create_user_with_empty_string_as_password(self):

        policy = {}
        user = "user5-test"
        password = ""
        roles = ["read-write"]

        try:
            self.client.admin_drop_user(user, policy)
            time.sleep(2)
        except Exception:
            pass

        status = self.client.admin_create_user(user, password, roles, policy)

        assert status == 0
        time.sleep(2)
        self.delete_users.append(user)

    def test_create_user_with_special_characters_in_password(self):

        policy = {}
        user = "user6-test"
        password = "@#!$#$WERWE%&%$"
        roles = ["sys-admin"]

        try:
            self.client.admin_drop_user(user, policy)
            time.sleep(2)
        except Exception:
            pass

        status = self.client.admin_create_user(user, password, roles, policy)

        assert status == 0

        self.delete_users.append(user)

    def test_create_user_with_too_long_username(self):

        policy = {}
        user = "user$" * 1000
        password = "user10-test"
        roles = ["sys-admin"]

        try:
            self.client.admin_drop_user(user, policy)
            time.sleep(2)
        except Exception:
            pass

        try:
            self.client.admin_create_user(user, password, roles, policy)

        except e.InvalidUser as exception:
            assert exception.code == 60
            assert exception.msg == "AEROSPIKE_INVALID_USER"

        except e.ClientError:
            pass

    def test_create_user_with_too_long_password(self):

        policy = {"timeout": 1000}
        user = "user10-test"
        password = "user#" * 1000
        roles = ["read-write"]

        try:
            self.client.admin_drop_user(user, policy)
            time.sleep(2)
        except Exception:
            pass

        with pytest.raises(e.ClientError):
            self.client.admin_create_user(user, password, roles, policy)

    def test_create_user_with_empty_roles_list(self):

        policy = {}
        user = "user7"
        password = "user7"
        roles = []

        try:
            self.client.admin_drop_user(user, policy)
            time.sleep(2)
        except Exception:
            pass

        try:
            self.client.admin_create_user(user, password, roles, policy)

        except e.InvalidRole as exception:
            assert exception.code == 70
            assert exception.msg == "AEROSPIKE_INVALID_ROLE"

    def test_create_user_with_non_user_admin_user(self):

        policy = {}
        user = "non_admin_test"
        password = "non_admin_test"
        roles = ["read-write"]

        try:
            self.client.admin_drop_user(user, policy)
            time.sleep(2)
        except Exception:
            pass

        status = self.client.admin_create_user(user, password, roles, policy)
        time.sleep(2)

        assert status == 0

        config = self.connection_config

        non_admin_client = None

        try:
            non_admin_client = aerospike.client(config).connect("non_admin_test", "non_admin_test")
            status = non_admin_client.admin_create_user("user78", password, roles, policy)

            if non_admin_client:
                non_admin_client.close()

        except e.RoleViolation as exception:
            assert exception.code == 81

        self.delete_users.append("non_admin_test")

    @pytest.mark.parametrize("roles", [{}, (), 5, "read-write"])
    def test_create_user_with_non_list_roles(self, roles):

        policy = {}
        user = "user7"
        password = "user7"
        try:
            self.client.admin_drop_user(user, policy)
        except Exception:
            pass

        with pytest.raises(e.ParamError):
            self.client.admin_create_user(user, password, roles)

    @pytest.mark.parametrize("list_item", [{}, (), 5, []])
    def test_create_user_with_invalid_roles_types(self, list_item):

        policy = {}
        user = "user7"
        password = "user7"
        roles = ["read-write", list_item]
        try:
            self.client.admin_drop_user(user, policy)
        except Exception:
            pass

        with pytest.raises(e.ClientError):
            self.client.admin_create_user(user, password, roles)

    def test_create_user_with_very_long_role_name(self):

        policy = {}
        user = "user7"
        password = "user7"
        roles = ["read-write", "abc" * 50]
        try:
            self.client.admin_drop_user(user, policy)
            time.sleep(2)
        except Exception:
            pass

        with pytest.raises(e.ClientError):
            self.client.admin_create_user(user, password, roles)



import aerospike
import marshal
import json

# Serializers and deserializers
# Both local and global serializers use json library
# Functions print which one is being used

def classSerializer(obj):
    print("Using class serializer")
    return json.dumps(obj)

def classDeserializer(bytes):
    print("Using class deserializer")
    return json.loads(bytes)

def localSerializer(obj):
    print("Using local serializer")
    return json.dumps(obj)

def localDeserializer(bytes):
    print("Using local deserializer")
    return json.loads(bytes)

# First client has class-level serializer set in aerospike module
aerospike.set_serializer(classSerializer)
aerospike.set_deserializer(classDeserializer)
config = {
    'hosts': [('127.0.0.1', 3000)]
}
client = aerospike.client(config).connect()

# Second client has instance-level serializer set in client config
config['serialization'] = (localSerializer, localDeserializer)
client2 = aerospike.client(config).connect()

# Keys: foo1, foo2, foo3
keys = [('test', 'demo', f'foo{i}') for i in range(1, 4)]
# Tuple is an unsupported type
tupleBin = {'bin': (1, 2, 3)}

# Use the default, built-in serialization (cPickle)
client.put(keys[0], tupleBin)
(_, _, bins) = client.get(keys[0])
print(bins)

# Expected output:
# {'bin': (1, 2, 3)}

# First client uses class-level, user-defined serialization
# No instance-level serializer was declared
client.put(keys[1], tupleBin, serializer=aerospike.SERIALIZER_USER)
(_, _, bins) = client.get(keys[1])
# Notice that json-encoding a tuple produces a list
print(bins)

# Expected output:
# Using class serializer
# Using class deserializer
# {'bin': [1, 2, 3]}

# Second client uses instance-level, user-defined serialization
# Instance-level serializer overrides class-level serializer
client2.put(keys[2], tupleBin, serializer=aerospike.SERIALIZER_USER)
(_, _, bins) = client2.get(keys[2])
print(bins)

# Expected output:
# Using local serializer
# Using local deserializer
# {'bin': [1, 2, 3]}

# Cleanup
client.batch_remove(keys)
client.close()
client2.close()



# -*- coding: utf-8 -*-
import pytest
import random
from datetime import datetime
from aerospike import exception as e
from aerospike_helpers.operations import operations as operation
from .test_base_class import TestBaseClass

import aerospike

random.seed(datetime.now())


class TestBitwiseOperations(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        """
        Setup Method
        """
        self.keys = []
        self.test_key = "test", "demo", "bitwise_op"
        self.true_bin = "true_bin"
        self.false_bin = "false_bin"

        self.as_connection.put(self.test_key, {self.true_bin: True, self.false_bin: False})
        self.keys.append(self.test_key)

        yield

        for key in self.keys:
            try:
                self.as_connection.remove(key)
            except e.AerospikeError:
                pass

    @pytest.mark.parametrize(
        "send_bool_as, expected_true, expected_false",
        [
            (aerospike.PY_BYTES, True, False),
            (aerospike.INTEGER, 1, 0),
            (aerospike.AS_BOOL, True, False),
            (100, True, False),
            (0, True, False),
            (-1, True, False),
        ],
    )
    def test_bool_read_write_pos(self, send_bool_as, expected_true, expected_false):
        """
        Write Python bools with different client configurations.
        """
        config = TestBaseClass.get_connection_config()
        config["send_bool_as"] = send_bool_as
        test_client = aerospike.client(config).connect(config["user"], config["password"])
        ops = [
            operation.write("cfg_true", True),
            operation.write("cfg_false", False),
        ]

        try:
            test_client.operate(self.test_key, ops)
        except e.InvalidRequest:
            if self.server_version < [5, 6] and send_bool_as == aerospike.AS_BOOL:
                pytest.mark.xfail(reason="Servers older than 5.6 do not support as_bool")
                pytest.xfail()

        _, _, bins = self.as_connection.get(self.test_key)
        test_client.close()
        #  We should not have changed the zeroes bin
        assert bins["cfg_true"] is expected_true
        assert bins["cfg_false"] is expected_false
        assert bins[self.true_bin] is True
        assert bins[self.false_bin] is False



# Adds record keys from a stream to a list
# But limits the number of keys to "lim"
def limit(lim: int, result: list):
    # Integers are immutable
    # so a list (mutable) is used for the counter
    c = [0]
    def key_add(record):
        key, metadata, bins = record
        if c[0] < lim:
            result.append(key)
            c[0] = c[0] + 1
        else:
            return False
    return key_add

from aerospike import predicates as p

keys = []
query.foreach(limit(2, keys))
print(len(keys)) # 2



# Check non-existent record
(key, meta) = client.exists(keyTuple)

print(key) # ('test', 'demo', 'key', bytearray(b'...'))
print(meta) # None

# Check existing record
client.put(keyTuple, {'bin1': 4})
(key, meta) = client.exists(keyTuple)

print(key) # ('test', 'demo', 'key', bytearray(b'...'))
print(meta) # {'ttl': 2592000, 'gen': 1}



# -*- coding: utf-8 -*-
import pytest
import aerospike
from aerospike import exception as e
from aerospike_helpers.operations import map_operations as map_ops
from .test_base_class import TestBaseClass

Server61 = True

if TestBaseClass.major_ver < 6 or (TestBaseClass.major_ver == 6 and TestBaseClass.minor_ver == 0):
    if pytest.__version__ < "3.0.0":
        Server61 = False
    else:
        Server61 = False


def get_map_result_from_operation(client, key, operations, res_bin):
    """
    Map operations return a 1 bin record.
    This operation extracts the dictionary from it.
    """
    _, _, result_bins = client.operate(key, operations)
    return result_bins[res_bin]


def maps_have_same_keys(map1, map2):
    """
    Return True iff the two maps have an identical set of keys
    """
    if len(map1) != len(map2):
        return False

    for key in map1:
        if key not in map2:
            return False
    return True


def maps_have_same_values(map1, map2):
    if len(map1) != len(map2):
        return False

    for key in map1:
        if key not in map2 or map1[key] != map2[key]:
            return False
    return True


def sort_map(client, test_key, test_bin):
    map_policy = {"map_write_mode": aerospike.MAP_CREATE_ONLY, "map_order": aerospike.MAP_KEY_ORDERED}
    operations = [map_ops.map_set_policy(test_bin, map_policy)]
    client.operate(test_key, operations)


class TestNewListOperationsHelpers(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        """
        Setup Method
        """
        self.keys = []
        self.test_map = {"a": 5, "b": 4, "c": 3, "d": 2, "e": 1, "f": True, "g": False}

        self.test_key = "test", "demo", "new_map_op"
        self.test_bin = "map"
        self.as_connection.put(self.test_key, {self.test_bin: self.test_map})
        self.keys.append(self.test_key)

        yield

        for key in self.keys:
            try:
                self.as_connection.remove(key)
            except e.AerospikeError:
                pass

    def test_map_set_policy(self):
        """
        Test setting map policy with an operation
        """
        map_policy = {"map_write_mode": aerospike.MAP_CREATE_ONLY, "map_order": aerospike.MAP_KEY_VALUE_ORDERED}
        operations = [map_ops.map_set_policy(self.test_bin, map_policy)]

        self.as_connection.operate(self.test_key, operations)

    def test_map_put(self):
        operations = [map_ops.map_put(self.test_bin, "new", "map_put")]
        self.as_connection.operate(self.test_key, operations)
        res_map = self.as_connection.get(self.test_key)[2][self.test_bin]
        assert res_map["new"] == "map_put"

    def test_map_put_items(self):
        new_items = {"new": 1, "new2": 3}
        expected_items = {"a": 5, "b": 4, "c": 3, "d": 2, "e": 1, "f": True, "g": False, "new": 1, "new2": 3}
        operations = [map_ops.map_put_items(self.test_bin, new_items)]
        self.as_connection.operate(self.test_key, operations)
        res_map = self.as_connection.get(self.test_key)[2][self.test_bin]

        assert res_map == expected_items

    def test_map_increment(self):
        operations = [map_ops.map_increment(self.test_bin, "a", 100)]
        self.as_connection.operate(self.test_key, operations)
        res_map = self.as_connection.get(self.test_key)[2][self.test_bin]
        assert res_map["a"] == self.test_map["a"] + 100

    def test_map_decrement(self):
        operations = [map_ops.map_decrement(self.test_bin, "a", 100)]
        self.as_connection.operate(self.test_key, operations)
        res_map = self.as_connection.get(self.test_key)[2][self.test_bin]
        assert res_map["a"] == self.test_map["a"] - 100

    def test_map_size(self):
        operations = [map_ops.map_size(self.test_bin)]
        assert get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin) == len(
            self.test_map
        )

    def test_map_clear(self):
        operations = [map_ops.map_clear(self.test_bin)]
        self.as_connection.operate(self.test_key, operations)
        res_map = self.as_connection.get(self.test_key)[2][self.test_bin]
        assert res_map == {}

    def test_map_remove_by_key(self):
        operations = [map_ops.map_remove_by_key(self.test_bin, "a", return_type=aerospike.MAP_RETURN_VALUE)]
        assert (
            get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)
            == self.test_map["a"]
        )

        res_map = self.as_connection.get(self.test_key)[2][self.test_bin]
        assert "a" not in res_map

    def test_map_remove_by_key_list(self):
        operations = [
            map_ops.map_remove_by_key_list(self.test_bin, ["a", "b", "c"], return_type=aerospike.MAP_RETURN_VALUE)
        ]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)

        assert set(ret_vals) == set([self.test_map["a"], self.test_map["b"], self.test_map["c"]])

        res_map = self.as_connection.get(self.test_key)[2][self.test_bin]
        assert "a" not in res_map
        assert "b" not in res_map
        assert "c" not in res_map

    def test_map_remove_by_key_range(self):
        operations = [map_ops.map_remove_by_key_range(self.test_bin, "a", "d", return_type=aerospike.MAP_RETURN_VALUE)]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)

        assert set(ret_vals) == set([self.test_map["a"], self.test_map["b"], self.test_map["c"]])

        res_map = self.as_connection.get(self.test_key)[2][self.test_bin]
        assert "a" not in res_map
        assert "b" not in res_map
        assert "c" not in res_map

    def test_map_remove_by_value(self):
        operations = [
            map_ops.map_remove_by_value(self.test_bin, self.test_map["a"], return_type=aerospike.MAP_RETURN_KEY)
        ]
        assert get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin) == ["a"]

        res_map = self.as_connection.get(self.test_key)[2][self.test_bin]
        assert "a" not in res_map

    def test_map_remove_by_value_list(self):
        operations = [
            map_ops.map_remove_by_value_list(
                self.test_bin,
                [self.test_map["a"], self.test_map["b"], self.test_map["c"]],
                return_type=aerospike.MAP_RETURN_KEY,
            )
        ]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)

        assert set(ret_vals) == set(["a", "b", "c"])

        res_map = self.as_connection.get(self.test_key)[2][self.test_bin]
        assert "a" not in res_map
        assert "b" not in res_map
        assert "c" not in res_map

    def test_map_remove_by_value_range(self):
        operations = [map_ops.map_remove_by_value_range(self.test_bin, 1, 4, return_type=aerospike.MAP_RETURN_KEY)]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)

        assert set(ret_vals) == set(["e", "d", "c"])

        res_map = self.as_connection.get(self.test_key)[2][self.test_bin]
        assert "e" not in res_map
        assert "d" not in res_map
        assert "c" not in res_map

    def test_map_remove_by_index(self):
        sort_map(self.as_connection, self.test_key, self.test_bin)
        operations = [map_ops.map_remove_by_index(self.test_bin, 1, return_type=aerospike.MAP_RETURN_KEY)]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)
        assert ret_vals == "b"
        res_map = self.as_connection.get(self.test_key)[2][self.test_bin]
        assert "b" not in res_map

    def test_map_remove_by_index_range(self):
        sort_map(self.as_connection, self.test_key, self.test_bin)
        operations = [map_ops.map_remove_by_index_range(self.test_bin, 1, 2, return_type=aerospike.MAP_RETURN_KEY)]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)
        assert ret_vals == ["b", "c"]
        res_map = self.as_connection.get(self.test_key)[2][self.test_bin]
        assert "b" not in res_map
        assert "c" not in res_map

    def test_map_remove_by_rank(self):
        operations = [map_ops.map_remove_by_rank(self.test_bin, 1, return_type=aerospike.MAP_RETURN_KEY)]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)
        assert ret_vals == "d"
        res_map = self.as_connection.get(self.test_key)[2][self.test_bin]
        assert "d" not in res_map

    def test_map_remove_by_rank_range(self):
        operations = [map_ops.map_remove_by_rank_range(self.test_bin, 1, 2, return_type=aerospike.MAP_RETURN_KEY)]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)
        assert set(ret_vals) == set(["d", "c"])
        res_map = self.as_connection.get(self.test_key)[2][self.test_bin]
        assert "d" not in res_map
        assert "c" not in res_map

    def test_map_get_by_key(self):
        operations = [map_ops.map_get_by_key(self.test_bin, "a", return_type=aerospike.MAP_RETURN_VALUE)]
        assert (
            get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)
            == self.test_map["a"]
        )

    def test_map_get_by_key_list(self):
        operations = [
            map_ops.map_get_by_key_list(self.test_bin, ["a", "b", "c"], return_type=aerospike.MAP_RETURN_VALUE)
        ]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)

        assert set(ret_vals) == set([self.test_map["a"], self.test_map["b"], self.test_map["c"]])

    def test_map_get_by_key_range(self):
        operations = [map_ops.map_get_by_key_range(self.test_bin, "a", "d", return_type=aerospike.MAP_RETURN_VALUE)]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)

        assert set(ret_vals) == set([self.test_map["a"], self.test_map["b"], self.test_map["c"]])

    def test_map_get_by_value(self):
        operations = [map_ops.map_get_by_value(self.test_bin, self.test_map["a"], return_type=aerospike.MAP_RETURN_KEY)]
        assert get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin) == ["a"]

    def test_map_get_by_value_list(self):
        operations = [
            map_ops.map_get_by_value_list(
                self.test_bin,
                [self.test_map["a"], self.test_map["b"], self.test_map["c"]],
                return_type=aerospike.MAP_RETURN_KEY,
            )
        ]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)

        assert set(ret_vals) == set(["a", "b", "c"])

    def test_map_get_by_value_range(self):
        operations = [map_ops.map_get_by_value_range(self.test_bin, 1, 4, return_type=aerospike.MAP_RETURN_KEY)]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)

        assert set(ret_vals) == set(["e", "d", "c"])

    def test_map_get_by_index(self):
        sort_map(self.as_connection, self.test_key, self.test_bin)
        operations = [map_ops.map_get_by_index(self.test_bin, 1, return_type=aerospike.MAP_RETURN_KEY)]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)
        assert ret_vals == "b"

    def test_map_get_by_index_range(self):
        sort_map(self.as_connection, self.test_key, self.test_bin)
        operations = [map_ops.map_get_by_index_range(self.test_bin, 1, 2, return_type=aerospike.MAP_RETURN_KEY)]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)
        assert ret_vals == ["b", "c"]

    def test_map_get_by_rank(self):
        operations = [map_ops.map_get_by_rank(self.test_bin, 1, return_type=aerospike.MAP_RETURN_KEY)]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)
        assert ret_vals == "d"

    def test_map_get_by_rank_range(self):
        sort_map(self.as_connection, self.test_key, self.test_bin)
        operations = [map_ops.map_get_by_rank_range(self.test_bin, 1, 2, return_type=aerospike.MAP_RETURN_KEY)]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)
        assert ret_vals == ["d", "c"]

    def test_map_get_exists_by_key_list(self):
        if not Server61:
            pytest.skip("It only applies to >= 6.1 enterprise edition")
        operations = [
            map_ops.map_get_by_key_list(self.test_bin, ["a", "b", "c"], return_type=aerospike.MAP_RETURN_EXISTS)
        ]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)

        assert ret_vals is True

    def test_map_get_exists_by_key_range(self):
        if not Server61:
            pytest.skip("It only applies to >= 6.1 enterprise edition")
        operations = [map_ops.map_get_by_key_range(self.test_bin, "a", "d", return_type=aerospike.MAP_RETURN_EXISTS)]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)

        assert ret_vals is True

    def test_map_get_exists_by_value(self):
        if not Server61:
            pytest.skip("It only applies to >= 6.1 enterprise edition")
        operations = [
            map_ops.map_get_by_value(self.test_bin, self.test_map["a"], return_type=aerospike.MAP_RETURN_EXISTS)
        ]
        assert get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin) is True

    def test_map_get_exists_by_value_list(self):
        if not Server61:
            pytest.skip("It only applies to >= 6.1 enterprise edition")
        operations = [
            map_ops.map_get_by_value_list(
                self.test_bin,
                [self.test_map["a"], self.test_map["b"], self.test_map["c"]],
                return_type=aerospike.MAP_RETURN_EXISTS,
            )
        ]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)

        assert ret_vals is True

    def test_map_get_exists_by_value_range(self):
        if not Server61:
            pytest.skip("It only applies to >= 6.1 enterprise edition")
        operations = [map_ops.map_get_by_value_range(self.test_bin, 1, 4, return_type=aerospike.MAP_RETURN_EXISTS)]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)

        assert ret_vals is True

    def test_map_get_exists_by_index(self):
        if not Server61:
            pytest.skip("It only applies to >= 6.1 enterprise edition")
        sort_map(self.as_connection, self.test_key, self.test_bin)
        operations = [map_ops.map_get_by_index(self.test_bin, 1, return_type=aerospike.MAP_RETURN_EXISTS)]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)
        assert ret_vals is True

    def test_map_get_exists_by_index_range(self):
        if not Server61:
            pytest.skip("It only applies to >= 6.1 enterprise edition")
        sort_map(self.as_connection, self.test_key, self.test_bin)
        operations = [map_ops.map_get_by_index_range(self.test_bin, 1, 2, return_type=aerospike.MAP_RETURN_EXISTS)]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)
        assert ret_vals is True

    def test_map_get_exists_by_rank(self):
        if not Server61:
            pytest.skip("It only applies to >= 6.1 enterprise edition")
        operations = [map_ops.map_get_by_rank(self.test_bin, 1, return_type=aerospike.MAP_RETURN_EXISTS)]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)
        assert ret_vals is True

    def test_map_get_exists_by_rank_range(self):
        if not Server61:
            pytest.skip("It only applies to >= 6.1 enterprise edition")
        sort_map(self.as_connection, self.test_key, self.test_bin)
        operations = [map_ops.map_get_by_rank_range(self.test_bin, 1, 2, return_type=aerospike.MAP_RETURN_EXISTS)]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)
        assert ret_vals is True



# -*- coding: utf-8 -*-

import pytest
from aerospike import exception as e

import aerospike


@pytest.mark.xfail(reason="Method is deprecated")
@pytest.mark.usefixtures("as_connection", "connection_config")
class TestInfo(object):
    def test_info_for_statistics(self):
        request = "statistics"
        hosts = [host for host in self.connection_config["hosts"]]
        nodes_info = self.as_connection.info(request, hosts)

        assert nodes_info is not None

        assert type(nodes_info) == dict

    @pytest.mark.parametrize(
        "container_type, container_name",
        [("namespaces", "test"), ("sets", "demo"), ("bins", "names")],
        ids=("namespace", "sets", "bins"),
    )
    def test_positive_info(self, container_type, container_name):
        """
        Test to see whether a namespace, set,
        and bin exist after a key is added
        """
        key = ("test", "demo", "list_key")
        rec = {"names": ["John", "Marlen", "Steve"]}

        self.as_connection.put(key, rec)
        hosts = [host for host in self.connection_config["hosts"]]

        response = self.as_connection.info(container_type, hosts)
        self.as_connection.remove(key)
        found = False
        for keys in response.keys():
            for value in response[keys]:
                if value is not None:
                    if container_name in value:
                        found = True

        assert found

    def test_info_with_config_for_statistics_and_policy(self):

        request = "statistics"
        policy = {"timeout": 1000}
        hosts = [host for host in self.connection_config["hosts"]]

        nodes_info = self.as_connection.info(request, hosts, policy)

        assert nodes_info is not None
        assert isinstance(nodes_info, dict)

    def test_info_for_invalid_request(self):

        request = "no_info"
        hosts = [host for host in self.connection_config["hosts"]]
        nodes_info = self.as_connection.info(request, hosts)

        assert isinstance(nodes_info, dict)
        assert nodes_info.values() is not None

    def test_info_with_none_request(self):
        """
        Test that sending None as the request raises an error
        """
        request = None
        hosts = [host for host in self.connection_config["hosts"]]

        try:
            self.as_connection.info(request, hosts)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Request must be a string"

    def test_info_without_parameters(self):

        with pytest.raises(TypeError) as err_info:
            self.as_connection.info()

        assert "argument 'command' (pos 1)" in str(err_info.value)

    def test_info_positive_for_sets_without_connection(self):
        """
        Test info positive for sets without connection
        """
        client1 = aerospike.client(self.connection_config)
        client1.close()

        with pytest.raises(e.ClusterError) as err_info:
            client1.info("sets", self.connection_config["hosts"])

        assert err_info.value.code == 11
        assert err_info.value.msg == "No connection to aerospike cluster"

    @pytest.mark.parametrize(
        "host_arg",
        [
            None,
            5,
            False,
            (),
            {},
        ],
        ids=["None", "int", "bool", "tuple", "dict"],
    )
    def test_info_incorrect_host_type(self, host_arg):
        request = "statistics"

        with pytest.raises(e.ParamError):
            self.as_connection.info(request, host_arg)

    def test_info_with_host_as_tuple(self):
        request = "statistics"
        hosts_tuple = tuple(self.connection_config["hosts"])
        with pytest.raises(e.ParamError):
            self.as_connection.info(request, hosts_tuple)

    @pytest.mark.skip("This returns an empty dict" + "unsure if this is correct behavior")
    def test_info_host_as_list_of_list(self):
        """
        This currently segfaults
        """
        request = "statistics"
        hosts_l_o_l = [["127.0.0.1", 3000]]

        nodes_info = self.as_connection.info(request, hosts_l_o_l)

        assert nodes_info is not None
        assert isinstance(nodes_info, dict)

    def test_incorrect_host_address_type(self):

        request = "statistics"

        config = [(127, 3000)]

        with pytest.raises(e.ParamError):
            self.as_connection.info(request, config)

    def test_host_address_too_long(self):
        """
        due to > 3.0.0 implementation, this should just result in an exception
        not a param error
        """
        request = "statistics"
        addr = "1" * 47  # longest possible ipv6 is 45 characters
        # longest port is 5 characters
        # we are using a 4 char port, so we add 2
        with pytest.raises(Exception):
            self.as_connection.info(request, [(addr, 3000)])
